# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::Client;

use strict;

use FileHandle;
use Carp;
use File::stat;
use IO::Socket;
use POSIX qw(getuid getgid getgroups);

use Katie::PerlRPC;
use Katie::NFSerrno;
use Katie::Errno;
use Katie::Constants;

my $VERSION = '0.06';

my($katie_server_info_fn) = '.katie-server-info';

my($user_info);

$Katie::KELEMENT_DIR = 1;
$Katie::KELEMENT_REG = 2;

sub new {
  shift(@_);
  my(%args) = @_;
  my($self);

  # if hostname is given, that's all we can find out
  if (exists($args{hostname})) {
    $self->{hostname} = $args{hostname};

  # if a directory is given, look in the info file
  } elsif (exists($args{dir})) {
    if (!defined($args{dir}) or $args{dir} eq '') {
      $args{dir} = '.';
    }

    $self = get_katie_server_info($args{dir});
    if (!($self->{hostname})) {
      return (undef, "couldn't get server info");
    }

  # no hostname, no directory -- abort
  } else {
    return (undef, 'no hostname specified');
  }

  if (!($self->{connection} = katie_connect($self->{hostname}))) {
    return (undef, "couldn't connect: $!");
  }

  return bless $self;
}

sub get_katie_server_info {
  my($dir) = @_;
  my($f) = new FileHandle;
  my(%result);

  if (!($f->open("$dir/$katie_server_info_fn", '<'))) {
    return undef;
  }

  chomp($result{hostname} = <$f>);
  chomp($result{view_id} = <$f>);
  chomp($result{vob_id} = <$f>);
  chomp($result{dcookie} = <$f>);

  return \%result;
}

sub katie_connect {
  my($hostname) = @_;
  my($socket) = IO::Socket::INET->new(
    PeerAddr => $hostname,
    PeerPort => 52843,
    Proto    => 'tcp',
  );
  if (!$socket) {
    die "couldn't connect to katie server at $hostname: $!\n";
  }
  my($perlrpc) = Katie::PerlRPC->new($socket);

  return $perlrpc;
}

sub get_dir_inode_and_dcookie {
  my($dir) = @_;
  my($stat, $server_info);

  if (!($stat = stat($dir))) {
    # how can we do this nicely?
    croak("$0: stat($dir): $!");
  }

  $server_info = get_katie_server_info($dir);
  return ($stat->ino, $server_info->{dcookie});
}

sub katie_rpc_call {
  my($self, $function, @args) = @_;

  $self->{connection}->write([$function, $user_info, @args]);
}

sub mkvob {
  my($self) = shift;
  my($vob_name, $vob_path, $comment) = @_;

  $self->katie_rpc_call('mkvob', $vob_name, $vob_path, $comment);
  return $self->{connection}->read()->[0];
}

sub rmvob {
  my($self) = shift;
  my($vob_name) = @_;

  $self->katie_rpc_call('rmvob', $vob_name);
  return $self->{connection}->read()->[0];
}

sub mkview {
  my($self) = shift;
  my($vob_name, $view_name, $view_path) = @_;

  $self->katie_rpc_call('mkview', $vob_name, $view_name, $view_path);
  return $self->{connection}->read()->[0];
}

sub rmview {
  my($self) = shift;
  my($view_name, $force) = @_;

  $self->katie_rpc_call('rmview', $view_name, $force);
  return $self->{connection}->read()->[0];
}

sub checkout {
  my($self) = shift;
  my($dir, $file, $reserved, $branch, $ver, $comment) = @_;
  my($dir_inode, $dir_dcookie);

  croak("$0: no view ID") unless exists($self->{view_id});

  ($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  $self->katie_rpc_call('checkout', $self->{view_id},
                        $dir_inode, $dir_dcookie, $file,
                        $reserved, $branch, $ver, $comment);
  return $self->{connection}->read()->[0];
}

# TODO: make $save_vupr and $from work
sub checkin {
  my($self) = shift;
  my($dir, $file, $preserve_mtime, $save_vupr,
     $identical, $from, $comment) = @_;
  my($dir_inode, $dir_dcookie);

  croak("$0: no view ID") unless exists($self->{view_id});

  ($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  $self->katie_rpc_call('checkin', $self->{view_id},
                        $dir_inode, $dir_dcookie, $file, $preserve_mtime,
                        $save_vupr, $identical, $from, $comment);
  return $self->{connection}->read()->[0];
}

sub rename_to_another {
  my($name, $suffix) = @_;

  if (!-e "$name.$suffix") {
    if (!rename($name, "$name.$suffix")) {
      warn "error renaming $name to $name.$suffix: $!\n";
      return NFSERR_IO;
    }
    return "$name.$suffix";
  }

  for my $count (1..9999) {
    if (!-e "$name.$suffix.$count") {
      if (!rename($name, "$name.$suffix.$count")) {
        warn "error renaming $name to $name.$suffix.$count: $!\n";
        return NFSERR_IO;
      }
      return "$name.$suffix.$count";
    }
  }

  return undef;
}

sub mkelem {
  my($self) = shift;
  my($dir, $file, $el_type, $do_checkin, $no_checkout, $comment) = @_;
  my($dir_inode, $dir_dcookie);
  my($renamed_name);
  my($result);

  croak("$0: no view ID") unless exists($self->{view_id});

  ($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  # $no_checkout overrides $do_checkin
  $do_checkin = 0 if ($do_checkin and $no_checkout);

  # can't checkin if file doesn't exist
  return NFSERR_NOENT if ($do_checkin and !-e "$dir/$file");

  # can't create a directory if something already exists with that name
  return KATIEERR_MKDIREXISTFILE if ($el_type == eDIR and -e "$dir/$file");

  # can't create a directory and checkin together
  return KATIEERR_MKDIRNOCI if ($el_type == eDIR and $do_checkin);

  # file exists -- rename it before calling mkelem
  if (-e "$dir/$file") {
    my($suffix);

    if ($no_checkout) {
      $suffix = "keep";
    } else {
      $suffix = "mkelem";
    }

    if (!defined($renamed_name = rename_to_another("$dir/$file", $suffix))) {
      # XXX better return code?
      return NFSERR_EXIST;
    }
  }

  $self->katie_rpc_call('mkelem', $self->{view_id},
                        $dir_inode, $dir_dcookie, $file, $el_type, $comment);
  $result = $self->{connection}->read()->[0];

  if ($result != 0) {
    if ($renamed_name) {
      rename($renamed_name, "$dir/$file");
    }
    return $result;
  }

  if ($no_checkout) {
    return 0;
  }

  $self->katie_rpc_call('checkout', $self->{view_id},
                        $dir_inode, $dir_dcookie, $file,
                        1, undef, undef, $comment);
  $result = $self->{connection}->read()->[0];

  if ($result != 0) {
    return $result;
  }

  if (defined($renamed_name) and !rename($renamed_name, "$dir/$file")) {
    warn "error renaming $renamed_name to $dir/$file: $!\n";
    return NFSERR_IO;
  }

  if ($do_checkin) {
    $self->katie_rpc_call('checkin', $self->{view_id},
                          $dir_inode, $dir_dcookie, $file, 1,
                          0, 0, undef, $comment);
    $result = $self->{connection}->read()->[0];
  }

  return $result;
}

sub rmname {
  my($self) = shift;
  my($dir, $file, $comment) = @_;
  my($dir_inode, $dir_dcookie);

  croak("$0: no view ID") unless exists($self->{view_id});

  ($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  $self->katie_rpc_call('rmname', $self->{view_id},
                        $dir_inode, $dir_dcookie, $file, $comment);
  return $self->{connection}->read()->[0];
}

sub mklbtype {
  my($self) = shift;
  my($name, $replace, $pbranch) = @_;

  croak("$0: no view ID") unless exists($self->{view_id});

  $self->katie_rpc_call('mklbtype', $self->{view_id},
                        $name, $replace, $pbranch);
  return $self->{connection}->read()->[0];
}

sub mklabel {
  my($self) = shift;
  my($replace, $recurse, $label_type, $dir, $file) = @_;
  my($dir_inode, $dir_dcookie);

  croak("$0: no view ID") unless exists($self->{view_id});

  ($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  $self->katie_rpc_call('mklabel', $self->{view_id},
                        $replace, $recurse, $label_type,
                        $dir_inode, $dir_dcookie, $file);
  return $self->{connection}->read()->[0];
}

sub mkbrtype {
  my($self) = shift;
  my($name, $replace, $pbranch, $comment) = @_;

  croak("$0: no view ID") unless exists($self->{view_id});

  $self->katie_rpc_call('mkbrtype', $self->{view_id},
                        $name, $replace, $pbranch, $comment);
  return $self->{connection}->read()->[0];
}

sub mkbranch {
  my($self) = shift;
  my($branch_type, $dir, $file, $comment) = @_;
  my($dir_inode, $dir_dcookie);

  croak("$0: no view ID") unless exists($self->{view_id});

  ($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  $self->katie_rpc_call('mkbranch', $self->{view_id},
                        $branch_type, $dir_inode, $dir_dcookie,
                        $file, $comment);
  return $self->{connection}->read()->[0];
}

sub lsvob {
  my($self, @vob_tag_list) = @_;
  my(@result);

  $self->katie_rpc_call('lsvob', @vob_tag_list);
  MSG: while (1) {
    my($result) = $self->{connection}->read();
    return 0 if (!$result);

    for my $vob_info (@$result) {
      last MSG if (!defined($vob_info));
      push(@result, $vob_info);
    }
  }

  return 1, @result;
}

sub lsview {
  my($self, $vob_id, @view_tag_list) = @_;
  my(@result);

  $self->katie_rpc_call('lsview', $vob_id, @view_tag_list);
  MSG: while (1) {
    my($result) = $self->{connection}->read();
    return 0 if (!$result);

    for my $view_info (@$result) {
      last MSG if (!defined($view_info));
      push(@result, $view_info);
    }
  }

  return 1, @result;
}

sub protect {
  my($self) = shift;
  my($owner, $group, $mode,
     $recurse, $only_file, $only_directory,
     $dir, $file) = @_;
  my($dir_inode, $dir_dcookie);

  croak("$0: no view ID") unless exists($self->{view_id});

  ($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  $self->katie_rpc_call('protect', $self->{view_id},
                        $owner, $group, $mode,
                        $recurse, $only_file, $only_directory,
                        $dir_inode, $dir_dcookie, $file);
  return $self->{connection}->read()->[0];
}

# set $current to true to do a setcs to the current cs
# set $config_spec to undef to set the default cs
sub setcs {
  my($self) = shift;
  my($current, $config_spec) = @_;
  my($stat);

  $config_spec = $self->getcs() if ($current);

  croak("$0: no view ID") unless exists($self->{view_id});

  $self->katie_rpc_call('setcs', $self->{view_id}, $config_spec);
  return $self->{connection}->read()->[0];
}

sub getcs {
  my($self) = shift;
  my($stat);

  croak("$0: no view ID") unless exists($self->{view_id});

  $self->katie_rpc_call('getcs', $self->{view_id});
  return $self->{connection}->read()->[0];
}

sub get_user_info {
  $user_info = {
    uid    => getuid(),
    gid    => getgid(),
    groups => [ getgroups() ],
    umask  => umask,
  };
}

sub ls {
  my($self) = shift;
  my($want_dir, $dir, $file) = @_;

  croak("$0: no view ID") unless exists($self->{view_id});

  my($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  $self->katie_rpc_call('ls', $self->{view_id},
                        $want_dir, $dir_inode, $dir_dcookie, $file);
}

sub lshistory {
  my($self) = shift;
  my($recurse, $want_dir, $checkout, $branch_name, $dir, $file) = @_;

  croak("$0: no view ID") unless exists($self->{view_id});

  my($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  $self->katie_rpc_call('lshistory', $self->{view_id},
                        $recurse, $want_dir, $checkout, $branch_name,
                        $dir_inode, $dir_dcookie, $file);
}

sub lsvtree {
  my($self) = shift;
  my($dir, $file, $branch_name,
     $nrecurse, $no_labels, $nco,
     $all_revisions, $all_labels) = @_;

  croak("$0: no view ID") unless exists($self->{view_id});

  my($dir_inode, $dir_dcookie) = get_dir_inode_and_dcookie($dir);

  $self->katie_rpc_call('lsvtree', $self->{view_id},
                        $dir_inode, $dir_dcookie, $file, $branch_name,
                        $nrecurse, $no_labels, $nco,
                        $all_revisions, $all_labels);
}

sub get_next {
  my($self) = shift;

  my($result) = $self->{connection}->read();
  die "read from server connection failed\n" unless $result;
  die "read from server was not a ref\n" unless ref $result;
  return undef if ref $result eq 'SCALAR';
  # this will go once we don't have return codes
  return undef if ref $result eq 'ARRAY';
  die "read from server was not a hash ref" unless ref $result eq 'HASH';
  return $result;
}

get_user_info();

1;
