# Katie

# Copyright 2000-2002 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::ConfigSpec;

use strict;
use Carp qw(cluck);
use Time::Local qw(timegm timelocal timelocal_nocheck);
use Katie::VOB;
use Katie::Constants;
use Katie::File;

require Exporter;

@Katie::ConfigSpec::ISA = qw(Exporter);

@Katie::ConfigSpec::EXPORT = qw(
  parse_revision_selector
  set_cs
  get_cs_entry
  select_revision
  select_revision_id
  get_elbr_id_by_selector
);
# get_cs_mtime

# FIXME
sub get_inode_type {
  return Katie::View::get_inode_type(@_);
}

# Translate revision_selector text into a hash.  Returns a hash ref on
# success, undef on failure.
# Examples:
#
# /main/4
# { branch_path => [ ], label => '4' }
#
# MYLABEL
# { label => 'MYLABEL' }
#
# /main/foo/bar/5
# {
#   branch_path => ['foo', 'bar'],
#   label => '5'
# }
sub parse_revision_selector {
  my($selector_text) = @_;
  my(@branch_path, $label);

  if ($selector_text eq 'CHECKEDOUT') {
    return { label => $selector_text };
  }

  # Remove duplicate slashes
  $selector_text =~ s"/{2,}"/";

  # Change /LABEL to .../LABEL
  $selector_text = "...$selector_text" if $selector_text =~ m'^/[^/]+$';

  # TODO: check for /?{query}
  # if ($selector_text =~ m'^/?({.*})$') { ... }

  # If it doesn't begin with a slash or ".../", prepend ".../"
  $selector_text = ".../$selector_text" if $selector_text !~ m"^/|^\.\.\./";

  # Change .../main to /main
  $selector_text =~ s"^\.\.\./main/"/main/";

  # Change .../whatever to /main/.../whatever
  $selector_text =~ s"^\.\.\./"/main/.../";

  # Strip leading/trailing slashes
  $selector_text =~ s'^/+|/+$'';

  # Split on slashes into (@branch_path, $label)
  @branch_path = split(m'/+', $selector_text);
  $label = pop(@branch_path);

  # Don't allow bad labels
  return undef if !valid_label_name($label) and $label !~ /^[0-9]+$/;

  # Check that the first component is /main
  return undef unless shift @branch_path eq 'main';

  # Return no branch_path if it is just '...'
  if (@branch_path == 1 and $branch_path[0] eq '...') {
    return { label => $label };
  }

  return { branch_path => \@branch_path, label => $label };
}

sub check_date {
  my($y, $m, $d) = @_;

  return 0 unless $d >= 1 and $d <= 31;
  return 0 unless $m >= 1 and $m <= 12;
  return 0 unless $y >= 1970 and $y <= 2037;

  # Check for bad days like 31/Feb
  my($time) = timegm(0, 0, 0, $d, $m - 1, $y - 1900);
  my(undef, undef, undef, $d2, $m2, $y2) = gmtime($time);
  return ($d == $d2 and $m == $m2 and $y == $y2);
}

sub parse_date {
  my($date) = @_;
  my(%dow) = (
    'sun' => 0, 'sunday'    => 0,
    'mon' => 1, 'monday'    => 1,
    'tue' => 2, 'tuesday'   => 2,
    'wed' => 3, 'wednesday' => 3,
    'thu' => 4, 'thursday'  => 4,
    'fri' => 5, 'friday'    => 5,
    'sat' => 6, 'saturday'  => 6,
  );
  my(%moy) = (
    'jan' =>  1, 'january'   =>  1,
    'feb' =>  2, 'february'  =>  2,
    'mar' =>  3, 'march'     =>  3,
    'apr' =>  4, 'april'     =>  4,
    'may' =>  5,
    'jun' =>  6, 'june'      =>  6,
    'jul' =>  7, 'july'      =>  7,
    'aug' =>  8, 'august'    =>  8,
    'sep' =>  9, 'september' =>  9,
    'oct' => 10, 'october'   => 10,
    'nov' => 11, 'november', => 11,
    'dec' => 12, 'december', => 12,
  );

  $date = lc($date);

  # today
  if ($date eq 'today') {
    my(undef, undef, undef, $mday, $mon, $year) = localtime(time());
    return ($year + 1900, $mon + 1, $mday, '');
  }

  # yesterday
  if ($date eq 'yesterday') {
    my(undef, undef, undef, $mday, $mon, $year) = localtime(time() - 86400);
    return ($year + 1900, $mon + 1, $mday, '');
  }

  # day of week
  if (exists($dow{$date})) {
    my(undef, undef, undef, $mday, $mon, $year, $wday) = localtime(time());
    my($days_ago) = ($wday - $dow{$date}) % 7;
    (undef, undef, undef, $mday, $mon, $year) =
        localtime(time() - 86400 * $days_ago);
    return ($year + 1900, $mon + 1, $mday, 'week');
  }

  # dd-mm-yyyy or dd-mm
  if ($date =~ /^(\d+)-([a-z][a-z][a-z]|\d+)(-(\d\d\d\d))?$/i) {
    my($y, $m, $d) = ($4, $2, $1);
    my($rel) = '';

    # convert month name to number
    $m = lc($m);
    $m = $moy{$m} if exists($moy{$m});

    if (!defined($y)) {
      (undef, undef, undef, undef, undef, $y) = localtime(time());
      $y += 1900;
      $rel = 'year';
    }

    return undef unless check_date($y, $m, $d);
    return ($y, $m, $d, $rel);
  }

  # yyyy-mm-dd
  if ($date =~ /^(\d\d\d\d)-(\d\d)-(\d\d)$/) {
    my($y, $m, $d) = ($1, $2, $3);

    return undef unless check_date($y, $m, $d);
    return ($y, $m, $d, '');
  }

  # give up
  return undef;
}

sub check_time {
  my($h, $m, $s) = @_;

  return 0 unless $h >= 0 and $h <= 23;
  return 0 unless $m >= 0 and $m <= 60;
  return 0 unless $s >= 0 and $s <= 60;
  return 1;
}

sub parse_time {
  my($time) = @_;

  # hh:mm and hh:mm:ss
  if ($time =~ /^(\d+):(\d+)(:(\d+))?$/) {
    my($h, $m, $s) = ($1, $2, $4);

    $s = 0 unless defined($s);

    return undef unless check_time($h, $m, $s);
  } else {
    return undef;
  }
}

sub parse_datetime {
  my($datetime) = @_;
  my($date, $time);
  my($y, $m, $d, $rel);
  my($h, $min, $s);

  # check for now
  return time() if ($datetime eq 'now');

  # check for date.time
  if ($datetime =~ /^(.*)\.(.*)$/) {
    $date = $1;
    $time = $2;

    ($y, $m, $d, $rel) = parse_date($date);
    ($h, $min, $s) = parse_time($time);

    return undef unless defined($y) and defined($h);

    my($unix_time) = timelocal($s, $min, $h. $d, $m, $y);

    if (!$rel or $unix_time <= time()) {
      return $unix_time;
    }

    if ($rel == 'year') {
      $y--;
    } elsif ($rel == 'week') {
      $d -= 7;
    }

    return timelocal_nocheck($s, $min, $h, $d, $m, $y);
  }

  # check for date
  ($y, $m, $d, $rel) = parse_date($datetime);
  return timelocal(0, 0, 0, $d, $m, $y) if defined($y);

  # check for time
  ($h, $min, $s) = parse_time($datetime);
  return undef unless defined($h);
  (undef, undef, undef, $d, $m, $y) = localtime(time());
  my($unix_time) = timelocal($s, $min, $h, $d, $m, $y);
  return $unix_time if $unix_time <= time();
  $d--;
  return timelocal_nocheck($s, $min, $h, $d, $m, $y);
}

sub parse_config_spec {
  my($cs_text) = @_;
  my($line_num) = 0;
  local($_);
  my(@cs);

  for (split(/\n/, $cs_text)) {
    $line_num++;
    s/^ +| +$//;
    next if /^$|^#/;

    my($cs_line);

    if (/^element\s+((-file|-directory)\s+)?(\S+)\s+(\S+)(\s+(.*))?$/) {
      my(@options);

      $cs_line->{scope} =
          !defined($2)
        ? 'all'
        : $2 eq '-file'
        ? 'file'
        : 'directory';

      $cs_line->{pattern} = $3;

      $cs_line->{selector} = parse_revision_selector($4);
      return (undef, $line_num) unless defined($cs_line->{selector});
      $cs_line->{selector}->{text} = $4;

      @options = split(/\s+/, $6) if defined($6);

      my($option);
      while ($option = shift @options) {
        if ($option eq '-time') {
          my($time) = shift @options;
          return (undef, $line_num) unless defined($time);
          $cs_line->{time} = parse_datetime($time);
          return (undef, $line_num) unless defined($cs_line->{time});
        } elsif ($option eq '-nocheckout') {
          $cs_line->{nocheckout} = 1;
        } elsif ($option eq '-mkbranch') {
          $cs_line->{mkbranch} = shift @options;
          return (undef, $line_num) unless defined($cs_line->{mkbranch});
        } else {
          return (undef, $line_num);
        }
      }

      push(@cs, $cs_line);
    } else {
      return (undef, $line_num);
    }
  }

  return (\@cs, undef);
}

{
  # config spec cache
  # $cs_cache{$view_id}->{spec}
  # $cs_cache{$view_id}->{mtime}
  # $cs_cache{$view_id}->{compiled}->[$cs_line_num]
  my(%cs_cache);

  my($built_in_default_cs) = {
    spec     => BUILT_IN_DEFAULT_CS,
    mtime    => 0,
    compiled => (parse_config_spec(BUILT_IN_DEFAULT_CS))[0],
  };

  # set a new config spec -- set $cs_text to undef to use the default
  # returns 0 for no error, otherwise the line number of the error
  sub set_cs {
    my($vob_db, $view_id, $cs_text) = @_;

    $cs_text = read_file(DEFAULT_CONFIG_SPEC_FN) if !defined($cs_text);
    if (!defined($cs_text)) {
      warn "couldn't read default config spec -- using built-in default\n";
      return $built_in_default_cs;
    }

    my($cs, $error_line) = parse_config_spec($cs_text);
    return ($error_line) if (!$cs);

    # put the new spec and mtime in the database and cache
    $vob_db->update("config_spec", $cs_cache{$view_id} = {
      spec  => $cs_text,
      mtime => time(),
    }, "view_id = $view_id");

    $cs_cache{$view_id}->{compiled} = $cs;

    return 0;
  }

  # return the cache entry from %cs_cache (if it's not there, create it)
  sub get_cs_entry {
    my($vob_db, $view_id) = @_;
    my($row);

    # check the cs_cache for it
    return $cs_cache{$view_id} if ($cs_cache{$view_id});

    # not in the cache -- check the database
    $row = $vob_db->select("
        SELECT spec, mtime
        FROM   config_spec
        WHERE  view_id = $view_id;
    ")->fetchrow();

    # not in the database -- return the built in default
    return $built_in_default_cs unless $row;

    # return the built in default if it fails to compile
    if (!($row->{compiled} = (parse_config_spec($row->{spec}))[0])) {
      warn "config spec for view $view_id failed to compile\n";
      return $built_in_default_cs;
    }

    return $cs_cache{$view_id} = $row;
  }
}

# TODO: make this work for normal shell patterns rather than for Perl
# regular expressions
sub pattern_matches_path {
  my($pattern, $path) = @_;

  $path = '/' if ($path eq '');
  return eval { $path =~ /$pattern/ };
}

sub follow_branch_path {
  my($vob_db, $element_branch_id, $branch) = @_;
  my($row);

  $branch = $vob_db->escape_sql($branch);

  $row = $vob_db->select("
      SELECT element_branch.id
      FROM   element_branch, branch, revision
      WHERE  branch.name = '$branch'
         AND branch.id = element_branch.branch_id
         AND element_branch.parent_revision_id = revision.id
         AND revision.element_branch_id = $element_branch_id;
  ")->fetchrow();

  return undef unless $row;
  return $row->{id};
}

sub get_branch_tree {
  my($vob_db, $element_branch_id) = @_;
  my(@branch_list, @branches_to_check);
  my($branch, $row);
  my($branch_info);

  $branch_info = (['', $element_branch_id]);
  push(@branch_list, $branch_info);
  push(@branches_to_check, $branch_info);

  while ($branch = shift @branches_to_check) {
    my($query) = $vob_db->select("
        SELECT branch.name, element_branch.id
        FROM   branch, element_branch, revision
        WHERE  branch.id = element_branch.branch_id
           AND element_branch.parent_revision_id = revision.id
           AND revision.element_branch_id = $branch->[1];
    ");
    while ($row = $query->fetchrow()) {
      my($branch_name) = "$branch->[0]/$row->{name}";
      my($branch_info) = [ $branch_name, $row->{id} ];
      push(@branch_list, $branch_info);
      push(@branches_to_check, $branch_info);
    }
  }

  return @branch_list;
}

sub get_branch_matches {
  my($vob_db, $element_branch_id, $branch_path) = @_;
  my($branch);
  my(@branch_path) = @$branch_path;

  # follow the branch path down -- abort if we hit '...'
  while ($branch = shift @branch_path) {
    last if ($branch eq '...');
    $element_branch_id = follow_branch_path($vob_db,
                                            $element_branch_id, $branch);
    return () unless $element_branch_id;
  }

  # if we didn't hit '...', then there's just one match
  return ($element_branch_id) unless defined($branch) and $branch eq '...';

  my(@branch_list) = get_branch_tree($vob_db, $element_branch_id);
  my($regex) = (
                 join '',
                 map { '/' . ($_ eq '...' ? '.*' : quotemeta($_)) }
                 @branch_path
               ) . '$';
  my(@branch_matches);
  for my $branch_info (@branch_list) {
    if ($branch_info->[0] =~ /$regex/) {
      push(@branch_matches, $branch_info->[1]);
    }
  }
  return (@branch_matches);
}

sub select_revision_by_selector {
  my($view_id, $element_id, $selector) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($element_branch_id);
  my(@branch_matches, $revision_id);

  # check for /?LABEL
  if (!$selector->{branch_path}) {
    return get_id_by_label($vob_id, $vob_db, $element_id, $selector->{label});
  }

  # follow the branch_path to get the element_branch_id
  $element_branch_id = get_main_element_branch($view_id, $element_id);
  @branch_matches = get_branch_matches($vob_db, $element_branch_id,
                                       $selector->{branch_path});

  for $element_branch_id (@branch_matches) {

    my($revision_num) = $selector->{label};
    my($new_revision_id);

    # check for LATEST
    if ($selector->{label} eq 'LATEST') {
      $revision_num = get_latest_revision_num($vob_db, $element_branch_id);
    }

    # check for revision
    if ($revision_num =~ /^[0-9]$/) {
      $new_revision_id =
          get_id_by_revision($vob_id, $vob_db,
                             $element_branch_id, $revision_num);
      if ($new_revision_id) {
        return 'error' if $revision_id;
        $revision_id = $new_revision_id;
        next;
      }
    }

    # must be a label
    $new_revision_id =
        get_id_by_label_on_elbr($vob_id, $vob_db,
                                $element_branch_id, $revision_num);
    if ($new_revision_id) {
      return 'error' if $revision_id;
      $revision_id = $new_revision_id;
    }
  }

  return $revision_id;
}

# given an element ID, return the revision number and config spec line
# that currently matches the config spec for $view_id
sub select_revision {
  my($view_id, $element_id, $dpath) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($element);
  my($query, $row);
  my($cs) = get_cs_entry($vob_db, $view_id)->{compiled};

  if (get_inode_type($view_id, $element_id) != iELEM) {
    # the caller should do this test themselves
    warn "select_revision_id called for non-element!\n";
    return ($element_id, undef);
  }

  unless (defined($dpath)) {
    cluck "select_revision_id missing dpath";
    $dpath = '';
  }

  $element = $inode_cache[$vob_id]->{$element_id};

  for my $cs_entry (@$cs) {

    # check scope
    next if ($cs_entry->{scope} eq 'file' and $element->{type} == eDIR);
    next if ($cs_entry->{scope} eq 'directory' and $element->{type} == eREG);

    # check pattern
    next unless pattern_matches_path($cs_entry->{pattern}, $dpath);

    # check for CHECKEDOUT
    if ($cs_entry->{selector}->{label} eq 'CHECKEDOUT') {
      my($checkout_row) = is_checked_out($view_id, $element_id);
      next unless $checkout_row;
      return ($checkout_row->{id}, $cs_entry);
    }

    # check version selector
    my($revision_id) = select_revision_by_selector($view_id, $element_id,
                                                   $cs_entry->{selector});

    # try next rule if no revision found
    next unless $revision_id;

    # abort if error (eg. two revisions found)
    return (undef, 'error') if $revision_id eq 'error';

    # otherwise return the found revision
    return ($revision_id, $cs_entry);
  }

  # no matching rule
  return (undef, 'nomatch');
}

# given an element ID, return the inode of the revision selected by
# the current configuration spec of the view
sub select_revision_id {
  return (select_revision(@_))[0];
}

sub parse_branch_selector {
  my($selector_text) = @_;
  my(@branch_path);

  # Remove duplicate slashes
  $selector_text =~ s"/{2,}"/";

  # If it doesn't begin with a slash or ".../", prepend ".../"
  $selector_text = ".../$selector_text" if $selector_text !~ m"^/|^\.\.\./";

  # Change .../main to /main
  $selector_text =~ s"^\.\.\./main(/|$)"/main$1";

  # Change .../whatever to /main/.../whatever
  $selector_text =~ s"^\.\.\./"/main/.../";

  # Strip leading/trailing slashes
  $selector_text =~ s'^/+|/+$'';

  # Split on slashes into @branch_path
  @branch_path = split(m'/+', $selector_text);

  # Check that the first component is /main
  return undef unless shift @branch_path eq 'main';

  # Return no branch_path if it is just '...'
  if (@branch_path == 1 and $branch_path[0] eq '...') {
    return undef;
  }

  return \@branch_path;
}

# FIXME: this function isn't related to ConfigSpec.pm
sub get_elbr_id_by_selector {
  my($vob_db, $view_id, $element_id, $branch_name) = @_;

  my $branch_path = parse_branch_selector($branch_name);
  return undef unless $branch_path;

  my $element_branch_id = get_main_element_branch($view_id, $element_id);
  my @branch_matches = get_branch_matches($vob_db, $element_branch_id,
                                          $branch_path);
  # return all matches in array context
  return @branch_matches if wantarray;

  # return undef in scalar context unless there's exactly one match
  return undef unless @branch_matches == 1;
  return $branch_matches[0];
}

1;
