# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::NFS;

use strict;

use FileHandle;
use POSIX;

use Katie::PGq;
use Katie::NFSfh;
use Katie::NFSerrno;
use Katie::Errno;
use Katie::Constants;
use Katie::DB;
use Katie::VOB;
use Katie::View;
use Katie::ViewFile;
use Katie::ViewLookup;
use Katie::ViewReaddir;
use Katie::ConfigSpec;
use Katie::Permissions;

# the info returned in the .katie-server-info file
my($hostname);

# NFS file types
my($NFNON)  = 0; # non-file
my($NFREG)  = 1; # regular file
my($NFDIR)  = 2; # directory
my($NFBLK)  = 3; # block special
my($NFCHR)  = 4; # character special
my($NFLNK)  = 5; # symbolic link
my($NFSOCK) = 6; # unix domain sockets
my($NFBAD)  = 7; # unused
my($NFFIFO) = 8; # named pipe

sub get_katie_server_info {
  my($view_id, $vob_id, $dcookie) = @_;

  # this shouldn't happen once we have dcookies for @@ (sub)directories
  if (!defined($dcookie)) {
    $dcookie = "";
  }

  return "$hostname\n$view_id\n$vob_id\n$dcookie\n";
}

sub nfs_getattr {
  my($auth, $fh) = @_;
  my($view_id, $inode_id, $dcookie) = decode_filehandle($fh);
  my($vob_id);
  my($inode_type);
  my($element_id, $is_elem_or_revi_or_cout, $element_type, $revision_id);
  my($type, $mode, $nlink, $uid, $gid, $size, $blocksize);
  my($rdev, $blocks, $fsid, $fileid, $mtime);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};

  if ($inode_id == KATIE_SERVER_INFO_INODE) {
    my($now) = time();

    return ($NFREG, 0444, 1, 0, 0,
            length(get_katie_server_info($view_id, $vob_id, $dcookie)),
            1024, 0, 1, $view_id, KATIE_SERVER_INFO_INODE, $now, $now, $now);
  }

  $inode_type = get_inode_type($view_id, $inode_id);

  if (!$inode_type) {
    warn "nfs_getattr: " .
         "get_inode_type($view_id, $inode_id) didn't return a valid value\n";
    return NFSERR_IO;
  }

  if ($inode_type == iVUPR) {
    return view_private_get_attr($view_id, $inode_id);
  }

  $element_id = get_element_id($view_id, $inode_id);
  if (!$element_id) {
    return NFSERR_NOENT;
  }

  $is_elem_or_revi_or_cout = (
    $inode_type == iELEM or
    $inode_type == iREVI or
    $inode_type == iCOUT
  );

  $mode = $inode_cache[$vob_id]->{$element_id}->{mode};

  $element_type = $inode_cache[$vob_id]->{$element_id}->{type};

  if ($element_type == eDIR || !$is_elem_or_revi_or_cout) {
    $type = $NFDIR;
  } elsif ($element_type == eREG) {
    $type = $NFREG;
  } else {
    warn "getattr($view_id, $inode_id): type $element_type not valid\n";
    $type = $NFBAD;
  }

  $rdev = 0;
  $fsid = $view_id;

  # good number?
  $blocksize = 1024;

  # Copy the inode number in
  $fileid = $inode_id;

  # Set the link count
  # TODO: the nlink calculation here is obviously wrong & should be fixed
  if ($element_type == eDIR) {
    $nlink = 2;
  } else {
    $nlink = 1;
  }

  $uid = $inode_cache[$vob_id]->{$element_id}->{uid};
  $gid = $inode_cache[$vob_id]->{$element_id}->{gid};

  # set the mtime and size for files/directories seen as iELEM, iREVI,
  # or iCOUT
  if ($is_elem_or_revi_or_cout) {

    # get the revision ID if we're looking at an element
    if ($inode_type == iELEM) {
      $revision_id =
          select_revision_id($view_id, $inode_id,
                             get_dpath_by_dcookie($vob_id, $dcookie));

      # select_revision_id() should always return a valid ID
      if (!$revision_id) {
        warn "getattr($view_id, $inode_id) didn't find a revision\n";
        return NFSERR_NOENT;
      }

    } else {
      $revision_id = $inode_id;
    }

    # set the size and mtime for files
    if ($element_type == eREG) {
      if ($inode_type == iCOUT) {
        $size = 0;
        $mtime = $inode_cache[$vob_id]->{$revision_id}->{mtime};
      } else {
        $size = $inode_cache[$vob_id]->{$revision_id}->{size};
        if (!defined($size)) {
          warn "getattr($view_id, $inode_id) couldn't find the size\n";
          return NFSERR_NOENT;
        }
        $mtime = $inode_cache[$vob_id]->{$revision_id}->{mtime};
      }

    # set the size and mtime for directories
    } elsif ($element_type == eDIR) {

      $size = 1024;

      # looking at a particular directory revision?
      if ($inode_type == iREVI) {
        $mtime = $inode_cache[$vob_id]->{$inode_id}->{mtime};

      # looking at a directory through a view
      } else {

        if (get_inode_type($view_id, $revision_id) == iCOUT) {

          # use the greater of the dir_mtime and inode mtimes
          $mtime = dir_mtime_get($view_id, $inode_id);
          if (!defined($mtime) or
              $inode_cache[$vob_id]->{$inode_id}->{mtime} > $mtime) {
            $mtime = $inode_cache[$vob_id]->{$inode_id}->{mtime};
          }

        } else {

          # use the greater of the dir_mtime and revision mtimes
          $mtime = dir_mtime_get($view_id, $inode_id);
          if (!defined($mtime) or
              $inode_cache[$vob_id]->{$revision_id}->{mtime} > $mtime) {
            $mtime = $inode_cache[$vob_id]->{$revision_id}->{mtime};
          }
        }

        # if the config spec has been changed since mtime, update
        # mtime to the config spec mtime
        my($vob_db) = $vob_cache{$vob_id}->{db};
        my($cs_mtime) = get_cs_entry($vob_db, $view_id)->{mtime};
        $mtime = $cs_mtime if $cs_mtime > $mtime;
      }
    }

    if (!defined($mtime)) {
      warn "mtime not defined for inode $view_id/$revision_id/$inode_id\n";
      $mtime = 0;
    }

  # set the mtime and size for iATAT and iELBR directories
  } else {
    $mtime = $inode_cache[$vob_id]->{$element_id}->{mtime};
    $size = 1024;

    # also, add execute permissions to the iATAT and iELBR versions of
    # regular files
    if ($element_type == eREG) {
      $mode |= ($mode & 0444) >> 2;
    }
  }

  # Calculate the number of blocks occupied (assume 1K blocks)
  $blocks = int(($size + 1023) / 1024);

  return ($type, $mode, $nlink, $uid, $gid, $size, $blocksize,
          $rdev, $blocks, $fsid, $fileid, $mtime, $mtime, $mtime);
}

sub nfs_setattr {
  my($auth, $fh, $mode, $uid, $gid, $size, $atime, $mtime) = @_;
  my($view_id, $inode_id) = decode_filehandle($fh);
  my($vob_id);
  my($inode_type);
  my($error);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};

  $inode_type = get_inode_type($view_id, $inode_id);
  return NFSERR_ROFS unless $inode_type == iVUPR;

  if ($inode_cache[$vob_id]->{$inode_id}->{type} == eDIR && $size != -1) {
    return NFSERR_ISDIR;
  }

  my(@attrs) = view_private_get_attr($view_id, $inode_id);

  return NFSERR_IO if @attrs == 1;

  if ($mode != -1 or $uid != -1 or $gid != -1) {
    # must be owner to change mode or gid
    if ((($mode != -1 or $gid != -1) and
         !ensure_owner($auth, $attrs[3])) or

    # must be root or in group to change gid
        ($gid != -1 and !ensure_group($auth, $attrs[4])) or

    # must be root to change uid
        ($uid != -1 and $auth->{uid} != 0)) {
      return NFSERR_PERM;
    }

    if (!view_private_setmodes($view_id, $inode_id, $mode, $uid, $gid)) {
      return NFSERR_IO;
    }
  }

  if ($mtime != -1) {

    # only owner may set time to something apart from now
    if (!ensure_owner($auth, $attrs[3])) {
      $atime = $mtime = time();
    }

    # file must be writeable to touch it at all
    return NFSERR_PERM unless ensure_writeable($auth, @attrs);

    return NFSERR_IO unless view_private_settime($view_id, $inode_id,
                                                 $atime, $mtime);
  }

  if ($size != -1) {
    return NFSERR_PERM unless ensure_writeable($auth, @attrs);
    return NFSERR_IO unless view_private_setsize($view_id, $inode_id, $size);
  }

  return nfs_getattr($auth, $fh);
}

sub nfs_root {
  warn "nfs_root unimplemented";
  return 0;
}

sub nfs_lookup {
  my($auth, $dir, $name) = @_;
  my($view_id, $dir_inode_id, $dir_dcookie) = decode_filehandle($dir);
  my($vob_id, $dir_dpath);
  my($inode_id);
  my($dcookie, $fh, @attr);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  if ($name eq KATIE_SERVER_INFO_FN) {
    $fh = encode_filehandle($view_id, KATIE_SERVER_INFO_INODE, $dir_dcookie);
    return ($fh, nfs_getattr($auth, $fh));
  }

  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $inode_id = lookup_name($view_id, $dir_inode_id, $dir_dpath, $name);
  return NFSERR_NOENT unless $inode_id;

  if ($dir_dcookie) {
    return NFSERR_STALE unless defined($dir_dpath);
    $dcookie = get_dcookie_by_dpath($vob_id, "$dir_dpath/$name");
  }

  $fh = encode_filehandle($view_id, $inode_id, $dcookie);
  return NFSERR_IO unless length($fh) == 32;

  @attr = nfs_getattr($auth, $fh);
  return @attr if @attr == 1;

  return ($fh, @attr);
}

sub nfs_readlink {
  warn "nfs_readlink unimplemented";
  return NFSERR_IO;
}

sub nfs_read {
  my($auth, $file, $offset, $count) = @_;
  my($view_id, $inode_id, $dcookie) = decode_filehandle($file);
  my($inode_type, $element_id);
  my($data, $data_len);
  my(@attr);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  if ($inode_id == KATIE_SERVER_INFO_INODE) {
    my($vob_id) = $view_cache{$view_id}->{vob_id};
    return (substr(get_katie_server_info($view_id, $vob_id, $dcookie),
                   $offset, $count),
            nfs_getattr($auth, $file));
  }

  $inode_type = get_inode_type($view_id, $inode_id);
  if ($inode_type == iVUPR) {
    ($data_len, $data) =
        view_private_read($view_id, $inode_id, $offset, $count);
  } else {
    my($vob_id) = $view_cache{$view_id}->{vob_id};

    if ($inode_type == iATAT || $inode_type == iELBR) {
      warn "read of atat/elbr attempted\n";
      return NFSERR_ISDIR;
    }
    $element_id = ($inode_type == iELEM)
                  ? $inode_id
                  : get_element_id($view_id, $inode_id);
    if (!$element_id) {
      warn "nfs_read: element_id($view_id, $inode_id) undef\n";
      return NFSERR_IO;
    }
    if ($inode_cache[$vob_id]->{$element_id}->{type} != eREG) {
      return NFSERR_ISDIR;
      # XXX what about symlinks?  does the client or server handle them?
    }
    if ($inode_type == iELEM) {
      $inode_id = select_revision_id($view_id, $inode_id,
                                     get_dpath_by_dcookie($vob_id, $dcookie));
    }
    ($data_len, $data) =
        $vob_cache{$vob_id}->{clear_obj}->read(
            $element_id, $inode_id, $offset, $count
        );
  }

  # $data is the error if $data_len is undef
  return $data unless defined($data_len);

  @attr = nfs_getattr($auth, $file);
  return @attr if @attr == 1;

  return ($data, @attr);
}

sub nfs_write {
  my($auth, $file, $offset, $data) = @_;
  my($view_id, $inode_id) = decode_filehandle($file);
  my($vob_id);
  my($inode_type);
  my($result, @attr);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};

  $inode_type = get_inode_type($view_id, $inode_id);
  return NFSERR_ROFS unless $inode_type == iVUPR;
  return NFSERR_ISDIR if $inode_cache[$vob_id]->{$inode_id}->{type} == eDIR;

  $result = view_private_write($view_id, $inode_id, $offset, $data);
  return $result if ($result);
  return nfs_getattr($auth, $file);
}

sub nfs_create {
  my($auth, $dir, $name, $mode, $uid, $gid) = @_;
  my($view_id, $dir_inode_id, $dir_dcookie) = decode_filehandle($dir);
  my($vob_id, $dir_dpath);
  my($dir_type);
  my($error, $inode_id, $dcookie, $file);
  my(@attr);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if ($dir_type == dVIEW || $dir_type == dVUPR) {
    if ($dir_type == dVIEW &&
        lookup_name_view($view_id, $dir_inode_id, $dir_dpath, $name)) {
      return NFSERR_EXIST;
    } elsif ($dir_type == dVUPR &&
             lookup_name_vupr($view_id, $dir_inode_id, $name)) {
      return NFSERR_EXIST;
    }

    # set correct defaults
    # don't allow non-root to create files owned by others
    # don't allow non-root to create files group-owned by others
    $uid = $auth->{uid} if ($uid == -1 or $auth->{uid} != 0);
    $gid = $auth->{gid} if ($gid == -1 or !ensure_group($auth, $gid));

    # XXX we ignore the size, atime, and mtime; is that bad?
    ($error, $inode_id) =
        view_private_create($view_id, $dir_inode_id, eREG,
                            $name, $mode, $uid, $gid);
  } elsif (defined($dir_type)) {
    return NFSERR_ROFS;
  } else {
    return NFSERR_NOTDIR;
  }

  return $error if $error;

  if ($dir_dcookie) {
    return NFSERR_STALE unless defined($dir_dpath);
    $dcookie = get_dcookie_by_dpath($vob_id, "$dir_dpath/$name");
  }

  $file = encode_filehandle($view_id, $inode_id, $dcookie);
  @attr = nfs_getattr($auth, $file);
  return @attr if @attr == 1;
  return ($file, @attr);
}

sub nfs_remove {
  my($auth, $dir, $name) = @_;
  my($view_id, $dir_inode_id, $dir_dcookie) = decode_filehandle($dir);
  my($vob_id, $dir_dpath);
  my($dir_type, $file_id, $inode_type);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  # Check that we were passed a view or view-private directory
  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if ($dir_type == dVIEW) {
    $file_id = lookup_name_view($view_id, $dir_inode_id, $dir_dpath, $name);
  } elsif ($dir_type == dVUPR) {
    $file_id = lookup_name_vupr($view_id, $dir_inode_id, $name);
  } else {
    return NFSERR_ROFS;
  }

  return NFSERR_NOENT unless $file_id;

  return NFSERR_ISDIR if get_dir_type($view_id, $file_id);
  return NFSERR_ROFS if $inode_type_cache[$vob_id]->{$file_id} == iELEM;
  $inode_type = $inode_cache[$vob_id]->{$file_id}->{type};
  # XXX symlink
  return view_unlink($view_id, $dir_inode_id, $name, $file_id);
}

sub nfs_rename {
  my($auth, $dir1, $name1, $dir2, $name2) = @_;
  my($view_id1, $dir_id1, $dir_dcookie1) = decode_filehandle($dir1);
  my($view_id2, $dir_id2, $dir_dcookie2) = decode_filehandle($dir2);
  my($vob_id, $dir_dpath1, $dir_dpath2);
  my($dir_type);
  my($file_id1, $file_id2);
  my($inode_type1);

  if (!check_vobview_cache($view_id1)) {
    return NFSERR_STALE;
  }
  if (!check_vobview_cache($view_id2)) {
    return NFSERR_STALE;
  }

  if ($view_id1 != $view_id2) {
    return NFSERR_STALE;
  }

  $vob_id = $view_cache{$view_id1}->{vob_id};
  $dir_dpath1 = get_dpath_by_dcookie($vob_id, $dir_dcookie1);
  $dir_dpath2 = get_dpath_by_dcookie($vob_id, $dir_dcookie2);

  # can only move from a view or view-private directory
  $dir_type = get_dir_type($view_id1, $dir_id1);
  if ($dir_type != dVIEW and $dir_type != dVUPR) {
    return NFSERR_ROFS;
  }

  # check that the source file exists
  $file_id1 = lookup_name($view_id1, $dir_id1, $dir_dpath1, $name1);
  return NFSERR_NOENT unless $file_id1;

  # can only move view-private files
  $inode_type1 = get_inode_type($view_id1, $file_id1);
  return NFSERR_IO unless $inode_type1;
  return NFSERR_ROFS unless $inode_type1 == iVUPR;

  # can only move to a view or view-private directory
  $dir_type = get_dir_type($view_id2, $dir_id2);
  if ($dir_type != dVIEW and $dir_type != dVUPR) {
    return NFSERR_ROFS;
  }

  # see if the destination name already exists
  $file_id2 = lookup_name($view_id2, $dir_id2, $dir_dpath2, $name2);

  # yes, it exists
  if ($file_id2) {
    my($errno);

    # can't link on top of a directory
    return NFSERR_ISDIR if get_dir_type($view_id2, $file_id2);

    # can only link on top of a view-private file
    if ($inode_type_cache[$vob_id]->{$file_id2} != iVUPR) {
      return NFSERR_ROFS;
    }

    # remove the file we're linking to
    $errno = view_unlink($view_id2, $dir_id2, $name2, $file_id2);

    return $errno if ($errno);
  }

  # move the file to the new spot
  return view_rename($view_id1, $dir_id1, $name1, $dir_id2, $name2);
}

sub nfs_link {
  my($auth, $file1, $dir2, $name2) = @_;
  my($view_id1, $file_id1) = decode_filehandle($file1);
  my($view_id2, $dir_id2, $dir_dcookie2) = decode_filehandle($dir2);
  my($vob_id, $dir_dpath2);
  my($inode_type1);
  my($file_id2);
  my($dir_type);
  my($errno);

  return NFSERR_STALE unless check_vobview_cache($view_id1);
  return NFSERR_STALE unless check_vobview_cache($view_id2);

  return NFSERR_STALE unless $view_id1 == $view_id2;

  $vob_id = $view_cache{$view_id1}->{vob_id};
  $dir_dpath2 = get_dpath_by_dcookie($vob_id, $dir_dcookie2);

  # get the inode of the linked-from file in the inode_cache
  if (!($inode_type1 = get_inode_type($view_id1, $file_id1))) {
    warn "nfslink: get_inode_type($view_id1, $file_id1) failed\n";
    return NFSERR_IO;
  }

  # don't allow directory hard links
  if ($inode_cache[$vob_id]->{$file_id1}->{type} == eDIR) {
    return NFSERR_ISDIR;
  }

  # can only create links in view and view-private directories
  $dir_type = get_dir_type($view_id2, $dir_id2);
  if ($dir_type != dVIEW and $dir_type != dVUPR) {
    return NFSERR_ROFS;
  }

  # check if the file we're linking to already exists
  $file_id2 = lookup_name($view_id2, $dir_id2, $dir_dpath2, $name2);

  # yes, it exists
  if ($file_id2) {

    # can't link on top of a directory
    return NFSERR_ISDIR if get_dir_type($view_id2, $file_id2);

    # can only link on top of a view-private file
    if ($inode_type_cache[$vob_id]->{$file_id2} != iVUPR) {
      return NFSERR_ROFS;
    }

    # remove the file we're linking to
    $errno = view_unlink($view_id2, $dir_id2, $name2, $file_id2);

    return $errno if ($errno);
  }

  ($errno, undef) =
    view_private_create($view_id2, $dir_id2, eREG,
                        $name2, undef, undef, undef, $file_id1);

  return $errno;
}

sub nfs_symlink {
  warn "nfs_symlink unimplemented";
  return NFSERR_IO;
}

sub nfs_mkdir {
  my($auth, $dir, $name, $mode, $uid, $gid) = @_;
  my($view_id, $dir_inode_id, $dir_dcookie) = decode_filehandle($dir);
  my($vob_id, $dir_dpath);
  my($dir_type);
  my($error, $inode_id, $dcookie, $file);
  my(@attr);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if ($dir_type == dVIEW || $dir_type == dVUPR) {
    if ($dir_type == dVIEW &&
        lookup_name_view($view_id, $dir_inode_id, $dir_dpath, $name)) {
      return NFSERR_EXIST;
    } elsif ($dir_type == dVUPR &&
             lookup_name_vupr($view_id, $dir_inode_id, $name)) {
      return NFSERR_EXIST;
    }

    # set correct defaults for uid and gid
    # don't allow non-root to create files owned by others
    # don't allow non-root to create files group-owned by others
    $uid = $auth->{uid} if ($uid == -1 or $auth->{uid} != 0);
    $gid = $auth->{gid} if ($gid == -1 or !ensure_group($auth, $gid));

    # XXX we ignore the size, atime, and mtime; is that bad?
    ($error, $inode_id) =
        view_private_create($view_id, $dir_inode_id, eDIR,
                            $name, $mode, $uid, $gid);
  } elsif (defined($dir_type)) {
    return NFSERR_ROFS;
  } else {
    return NFSERR_NOTDIR;
  }

  return $error if $error;

  if ($dir_dcookie) {
    my($vob_id) = $view_cache{$view_id}->{vob_id};
    my($dir_dpath) = get_dpath_by_dcookie($vob_id, $dir_dcookie);
    if (!defined($dir_dpath)) {
      return NFSERR_STALE;
    }
    $dcookie = get_dcookie_by_dpath($vob_id, "$dir_dpath/$name");
  }

  $file = encode_filehandle($view_id, $inode_id, $dcookie);
  @attr = nfs_getattr($auth, $file);
  return @attr if @attr == 1;
  return ($file, @attr);
}

sub nfs_rmdir {
  my($auth, $dir, $name) = @_;
  my($view_id, $dir_inode_id, $dir_dcookie) = decode_filehandle($dir);
  my($vob_id, $vob_db);
  my($dir_dpath);
  my($dir_type, $subdir_id);
  my($subdir_type);
  my($row);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};

  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  # Check that $dir is a view or view-private directory and get the
  # inode number for $name into $subdir_id
  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if ($dir_type == dVIEW) {
    $subdir_id = lookup_name_view($view_id, $dir_inode_id, $dir_dpath, $name);
  } elsif ($dir_type == dVUPR) {
    $subdir_id = lookup_name_vupr($view_id, $dir_inode_id, $name);
  } else {
    return NFSERR_ROFS;
  }

  return NFSERR_NOENT unless $subdir_id;

  # Check that $subdir_id is a view-private directory
  $subdir_type = get_dir_type($view_id, $subdir_id);
  return NFSERR_NOTDIR unless $subdir_type;
  return NFSERR_ROFS unless $subdir_type == dVUPR;

  # Check that $subdir_id is empty (only two entries -- presumably '.'
  # and '..')
  $row = $vob_db->select("
      SELECT COUNT(*)
      FROM   dir_private_entry
      WHERE  view_id = $view_id
         AND dir_id = $subdir_id;
  ")->fetchrow();
  return NFSERR_NOTEMPTY if $row->{count} > 2;

  # Remove the '.' and '..' entries from dir_private_entry
  $vob_db->exec("
      DELETE FROM dir_private_entry
      WHERE  view_id = $view_id
         AND dir_id = $subdir_id;
  ");

  # Remove the cache entry for $subdir_id
  delete($inode_type_cache[$vob_id]->{$subdir_id});
  delete($inode_cache[$vob_id]->{$subdir_id});

  # Remove the entry for $subdir_id from $dir
  $vob_db->exec("
      DELETE FROM dir_private_entry
      WHERE  view_id = $view_id
         AND dir_id = $dir_inode_id
         AND file_id = $subdir_id;
  ");

  # Drop the link_count on the subdirectory
  $vob_db->exec("
      UPDATE view_private
      SET    link_count = link_count - 2
      WHERE  id = $subdir_id;
  ");

  # Delete the dir_mtime entry for $subdir_id
  $vob_db->exec("
      DELETE FROM dir_mtime
      WHERE  view_id = $view_id
         AND dir_id = $subdir_id;
  ");
  delete($dir_mtime_cache[$vob_id]->{$view_id}->{$subdir_id});

  # Touch $dir
  dir_mtime_set($view_id, $dir_inode_id);

  return 0;
}

sub nfs_readdir {
  my($auth, $dir, $cookie, $count) = @_;
  my($view_id, $dir_inode_id, $dir_dcookie) = decode_filehandle($dir);
  my($vob_id, $dir_dpath);
  my($error, $eof, @dirlist);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  ($error, $eof, @dirlist) =
      view_readdir($view_id, $dir_inode_id, $dir_dpath, $cookie, $count);

  return $error if $error;

  # we reverse and dereference the arrays for the benefit of SunRPCServer.pm
  @dirlist = map {$_->[0], $_->[1], $_->[2]} reverse @dirlist;

  return () unless @dirlist;
  return ($eof, @dirlist);
}

sub nfs_statfs {
  my($auth, $fh) = @_;

  return (8192, 1024, 1, 1, 1);
  # TODO: write some code to return the result of statfs on the VOB
  # (view?) directory
}

$hostname = (uname())[1];

1;
