# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::NFSfh;

# This module handles the NFS filehandles and dcookies.
#
# Format of a file handle (32 bytes) is:
#  4 bytes - view ID
#  4 bytes - inode ID
# 24 bytes - dcookie
#
# The dcookies (directory entry cookies) are used to map between file
# handles and pathnames so that select_revision() can get the pathname
# of the file it is working on.
#
# The root directory pathname is the empty string.  Each component
# will append "/<file_name>" to that empty string.

use strict;

use Carp qw(cluck);

require Exporter;
@Katie::NFSfh::ISA = qw(Exporter);

@Katie::NFSfh::EXPORT = qw(
  get_dcookie_by_dpath
  get_dpath_by_dcookie
  encode_filehandle
  decode_filehandle
);

use constant DCOOKIE_LEN     => 24;
use constant DCOOKIE_NYB_LEN => DCOOKIE_LEN * 2;
use constant NULL_DCOOKIE    => '0' x DCOOKIE_NYB_LEN;
use constant FH_PACK         => 'llH' . DCOOKIE_NYB_LEN;

use vars qw(
  %dcookie_by_dpath
  %dpath_by_dcookie
);

sub make_dcookie {
  my($vob_id) = @_;
  my($dcookie);

  do {
    $dcookie = '';
    for my $count (1..DCOOKIE_LEN) {
      $dcookie .= sprintf("%02x", int(rand 256));
    }
  } while ($dpath_by_dcookie{$vob_id}->{$dcookie});

  return $dcookie;
}

sub get_dcookie_by_dpath {
  my($vob_id, $dpath) = @_;
  my($dcookie);


  if ($dcookie_by_dpath{$vob_id}->{$dpath}) {
    return $dcookie_by_dpath{$vob_id}->{$dpath};
  }

  $dcookie = make_dcookie($vob_id);

  $dcookie_by_dpath{$vob_id}->{$dpath} = $dcookie;
  $dpath_by_dcookie{$vob_id}->{$dcookie} = $dpath;

  return $dcookie;
}

sub get_dpath_by_dcookie {
  my($vob_id, $dcookie) = @_;

  if (!defined($dcookie)) {
    warn "dcookie not defined\n";
    return undef;
  }

  if (defined($dpath_by_dcookie{$vob_id}->{$dcookie})) {
    return $dpath_by_dcookie{$vob_id}->{$dcookie};
  }

  return undef;
}

sub encode_filehandle {
  my($view_id, $inode_id, $dcookie) = @_;

  if (!defined($dcookie)) {
    $dcookie = NULL_DCOOKIE;
  }

  return pack(FH_PACK, $view_id, $inode_id, $dcookie);
}

sub decode_filehandle {
  my($file_handle) = @_;
  my($view_id, $inode_id, $dcookie) = unpack(FH_PACK, $file_handle);

  if ($dcookie eq NULL_DCOOKIE) {
    undef $dcookie;
  }

  return ($view_id, $inode_id, $dcookie);
}

1;
