# Copyright 2000 NetCraft Australia Pty Ltd

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::PGq;

use strict;
use Pg;
use Carp;

sub get_connect_options {
  my($fn) = @_;
  my(@options);
  local($_);

  if (!-e $fn) {
    warn "$0: get_connect_options: '$fn' doesn't exist\n";
    return "";
  }

  if (!(open(F, "<$fn"))) {
    warn "$0: get_connect_options: couldn't open '$fn' for reading: $!\n";
    return "";
  }

  while (<F>) {
    chomp;
    s/^\s+//;
    s/\s+$//;
    s/\s*=\s*/=/;
    next if /^$/;
    next if /^#/;
    push(@options, $_);
  }
  return join(" ", @options);
}

sub new {
  shift(@_);
  my($connect_options, $connect_options_file) = @_;

  my($self) = {};
  $self->{CONN} = Pg::connectdb($connect_options . " " .
                                get_connect_options($connect_options_file));
  if ($self->{CONN}->status != PGRES_CONNECTION_OK) {
    die($self->{CONN}->errorMessage);
  }
  bless($self);
  $self->exec("set DateStyle to 'ISO';");
  return $self;
}

sub exec {
  my($self) = shift;
  my($query) = @_;

  my($result) = $self->{CONN}->exec($query);
  my($rs) = $result->resultStatus;

  if ($rs == PGRES_COMMAND_OK) {
    my($cs) = $result->cmdStatus;
    if ($query =~ /^\s*(UPDATE|DELETE)\s/i) {
      $cs =~ s/^(UPDATE|DELETE) //;
      return $cs + 0;
    }
    if ($query =~ /^\s*INSERT\s/i) {
      $cs =~ s/^INSERT \d+ //;
      return $cs + 0;
    }
    return $cs;
  }
  confess("exec result is $rs trying to execute $query\n" .
          $self->{CONN}->errorMessage . "\n");
}

sub select {
  my($self) = shift;
  my($query) = @_;

  my($result) = $self->{CONN}->exec($query);
  my($rs) = $result->resultStatus;

  if ($rs == PGRES_TUPLES_OK) {
    my($newself) = {};

    $newself->{NUMFIELDS} = $result->nfields;
    for (0 .. $result->nfields - 1) {
      $newself->{FIELDNAME}->[$_] = $result->fname($_);
    }
    $newself->{RESULT} = $result;
    return bless($newself);
  }
  confess("exec result is $rs trying to execute $query\n" .
          $self->{CONN}->errorMessage . "\n");
}

sub ntuples {
  my($self) = shift;

  return $self->{RESULT}->ntuples;
}

sub fetchrow {
  my($self) = shift;
  my(@row, %row);

  if (!(@row = $self->{RESULT}->fetchrow())) {
    return undef;
  }
  for (0 .. $self->{NUMFIELDS} - 1) {
    $row{$self->{FIELDNAME}->[$_]} = $row[$_];
  }
  return \%row;
}

sub insert {
  my($self) = shift;
  my($table, $row) = @_;

  my($field_list, @values, $values);

  $field_list = join(", ", sort keys(%{$row}));

  @values = ();
  for (sort keys(%{$row})) {
    if (!defined($row->{$_})) {
      push(@values, "NULL");
    } else {
      push(@values, "'" . $self->escape_sql($row->{$_}) . "'");
    }
  }
  $values = join(", ", @values);

  $self->exec("INSERT INTO $table ($field_list) VALUES ($values);");
}

sub update {
  my($self) = shift;
  my($table, $row, $where) = @_;

  my($field_list, @values, $values);

  $field_list = join(", ", sort keys(%{$row}));

  @values = ();
  for (sort keys(%{$row})) {
    if (!defined($row->{$_})) {
      push(@values, "$_ = NULL");
    } else {
      push(@values, "$_ = '" . $self->escape_sql($row->{$_}) . "'");
    }
  }
  $values = join(", ", @values);

  $self->exec("UPDATE $table SET $values WHERE $where");
}

sub transaction_begin {
  my($self) = shift;

  $self->exec("BEGIN;");
  $self->exec("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE;");
}

sub transaction_commit {
  my($self) = shift;

  $self->exec("COMMIT;");
}

sub transaction_abort {
  my($self) = shift;

  $self->exec("ABORT;");
}

sub fetch_value {
  my($self) = shift;
  my($table, $key_field, $key_value, $return_field) = @_;
  my($query, $row);

  $key_value = $self->escape_sql($key_value);
  $query = $self->select("
      SELECT $return_field AS result
      FROM   $table
      WHERE  $key_field = '$key_value';
  ");
  $row = $query->fetchrow();
  if (!$row) {
    warn "fetch_value($table, $key_field, $key_value, $return_field) " .
         "didn't match\n";
    return undef;
  }
  return $row->{result};
}

sub get_seq_currval {
  my($self) = shift;
  my($seq_name) = @_;
  my($query, $row);

  $query = $self->select("
      SELECT currval('$seq_name');
  ");
  $row = $query->fetchrow();
  if (!$row) {
    return undef;
  }
  return $row->{currval};
}

sub escape_sql {
  my($self) = shift;
  my($value) = @_;

  $value =~ s/\\/\\\\/g;
  $value =~ s/'/\\'/g;
  $value =~ s/"/\\"/g;
  return $value;
}

1;
