# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::Server;

use strict;
use POSIX;
use Carp qw(cluck);
use Katie::Constants;
use Katie::VOB;
use Katie::ConfigSpec;
use Katie::File;
use Katie::View;
use Katie::ViewLookup;
use Katie::ViewReaddir;
use Katie::Permissions;
use Katie::DB;
use Katie::NFSfh;
use Katie::NFSerrno;
use Katie::Errno;

sub is_natural {
  local($_) = @_;

  return (defined and /^[0-9]+$/);
}

sub katie_mkvob {
  my($client, $user_info, $vob_name, $vob_path, $comment) = @_;
  my($error, $vob_id);

  if (!valid_vob_name($vob_name)) {
    return KATIEERR_NAMEINVAL;
  }
  return NFSERR_INVAL unless defined $comment;

  $katie_db->transaction_begin();

  # prevent anything else from modifying the vob table
  $katie_db->exec("LOCK TABLE vob IN SHARE ROW EXCLUSIVE MODE;");

  # check that a VOB by that name doesn't already exist
  if ($error = vob_exists($vob_name)) {
    $katie_db->transaction_abort();
    return $error;
  }

  # create the VOB database and store
  ($error, $vob_id) = do_mkvob($vob_name, $vob_path, $user_info, $comment);
  if ($error) {
    $katie_db->transaction_abort();
    return $error;
  }

  my($vob_db) = $vob_cache{$vob_id}->{db};

  # create the main branch
  mkbrtype($vob_db, 'main', $user_info->{uid}, $user_info->{gid}, 0, 0,
           'the main branch');

  # create the root directory element
  do_mkelem($vob_id, undef, undef, undef, eDIR, $user_info, '');

  $katie_db->transaction_commit();

  return 0;
}

sub katie_mkview {
  my($client, $user_info, $vob_name, $view_name, $view_path) = @_;
  my($query, $row);
  my($vob_id, $vob_db);

  if (!valid_vob_name($vob_name) or
      !valid_view_name($view_name)) {
    return KATIEERR_NAMEINVAL;
  }

  $katie_db->transaction_begin();
  $katie_db->exec("LOCK TABLE view IN SHARE ROW EXCLUSIVE MODE;");
  $katie_db->exec("LOCK TABLE vob IN SHARE ROW EXCLUSIVE MODE;");

  # Get the vob_id
  $query = $katie_db->select("
      SELECT id
      FROM   vob
      WHERE  name = '$vob_name';
  ");
  $row = $query->fetchrow();
  if (!$row) {
    $katie_db->transaction_abort();
    return KATIEERR_VOBNEXIST;
  }

  $vob_id = $row->{id};
  load_vob_cache($vob_id);
  $vob_db = $vob_cache{$vob_id}->{db};

  # Ensure that the view doesn't exist
  $query = $katie_db->select("
      SELECT *
      FROM   view
      WHERE  name = '$view_name';
  ");
  $row = $query->fetchrow();
  if ($row) {
    $katie_db->transaction_abort();
    return NFSERR_EXIST;
  }

  # Create the view store
  if (!mkdir($view_path, 0700)) {
    $katie_db->transaction_abort();
    return $! + 0;
  }

  # Create the view
  $katie_db->insert("view", {
    vob_id => $vob_id,
    name   => $view_name,
    store  => $view_path,
  });

  my($view_id) = $katie_db->get_seq_currval('view_id_seq');
  my($config_spec);

  $config_spec = read_file(DEFAULT_CONFIG_SPEC_FN);
  $config_spec = BUILT_IN_DEFAULT_CS if (!defined($config_spec));

  $vob_db->insert("config_spec", {
    view_id => $view_id,
    spec    => $config_spec,
  });

  $katie_db->transaction_commit();
  return 0;
}

sub katie_rmview {
  my($client, $user_info, $view_name, $force) = @_;
  my($query, $row);
  my($view_id);
  my($vob_id, $vob_db);

  if ($view_name !~ /^[a-z][a-z0-9_.-]*$/) {
    return KATIEERR_NAMEINVAL;
  }

  # Get the view ID
  $query = $katie_db->select("
      SELECT *
      FROM   view
      WHERE  name = '$view_name';
  ");
  $row = $query->fetchrow();
  if (!$row) {
    return KATIEERR_VIEWNEXIST;
  }
  $view_id = $row->{id};

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};

  # Remove any checked-out files if $force is set
  $query = $vob_db->select("
      SELECT COUNT(*)
      FROM   checkout
      WHERE  view_id = $view_id
  ");
  $row = $query->fetchrow();
  if ($row->{count} > 0) {
    if (!$force) {
      return NFSERR_NOTEMPTY;
    }
    $vob_db->exec("
        DELETE FROM checkout
        WHERE  view_id = $view_id;
    ");
  }

  # TODO: write some more code
  # - delete/unlink the view-private files

  $katie_db->exec("
      DELETE FROM view
      WHERE  id = $view_id;
  ");
  delete($view_cache{$view_id});

  return 0;
}

sub katie_rmvob {
  my($client, $user_info, $vob_name) = @_;
  my($query, $row);
  my($vob_id);

  if ($vob_name !~ /^[a-z][a-z0-9_.-]*$/) {
    return KATIEERR_NAMEINVAL;
  }

  # Get the VOB ID
  $query = $katie_db->select("
      SELECT *
      FROM   vob
      WHERE  name = '$vob_name';
  ");
  $row = $query->fetchrow();
  if (!$row) {
    return NFSERR_NOENT;
  }
  $vob_id = $row->{id};

  # Forcibly delete all the views associated with the VOB
  # (if the VOB's being deleted, it doesn't matter if there are
  # checked-out files)
  $query = $katie_db->select("
      SELECT *
      FROM   view
      WHERE  vob_id = $vob_id;
  ");
  while ($row = $query->fetchrow()) {
    my($error) = katie_rmview($client, $user_info, $row->{name}, 1);
    if ($error) {
      return $error;
    }
  }

  do_rmvob($vob_id);

  return 0;
}

sub determine_checkout_revision {
  my($view_id, $file_id, $element_id, $revision_id, $cs_entry, $dir_dpath, $branch, $ver, $uid) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($branch_id);
  my($inode_type) = get_inode_type($view_id, $file_id);

  # $branch is defined -- select the latest revision from that branch
  if (defined($branch)) {
    my($selector) = parse_revision_selector("$branch/LATEST");

    return (undef, undef, KATIEERR_INVALIDBRPNAME) unless $selector;
    $revision_id =
        select_revision_by_selector($view_id, $element_id, $selector);
    $branch_id = $inode_cache[$vob_id]->{$revision_id}->{element_branch_id};
    return (undef, undef, KATIEERR_NOREVSELECTED) unless $revision_id;
    return ($revision_id, $branch_id, undef);
  }

  # $ver is defined -- we checkout the revision that was specified
  if (defined($ver)) {
    if ($inode_type == iATAT or $inode_type == iELBR) {
      return (undef, undef, KATIEERR_NOTAREV);
    }
    if ($inode_type == iREVI) {
      $branch_id = $inode_cache[$vob_id]->{$file_id}->{element_branch_id};
      return ($file_id, $branch_id, undef);
    }
    $branch_id = $inode_cache[$vob_id]->{$revision_id}->{element_branch_id};
    return ($revision_id, $branch_id, undef);
  }

  # neither $branch nor $ver was defined

  # this error should be more like 'Not a revision or a branch'
  return (undef, undef, KATIEERR_NOTAREV) if $inode_type == iATAT;

  # a branch was specified -- checkout the latest revision on it
  if ($inode_type == iELBR) {
    my($revision_num) = get_latest_revision_num($vob_db, $file_id);
    $revision_id =
        get_id_by_revision($vob_id, $vob_db, $file_id, $revision_num);
    $branch_id = $inode_cache[$vob_id]->{$revision_id}->{element_branch_id};
    return ($revision_id, $branch_id, undef);
  }

  # do the auto-make-branch stuff if the view file was given, or if
  # the branch selected by the config spec is the same one the
  # revision given is on
  if ($inode_type == iELEM ||
      $inode_cache[$vob_id]->{$revision_id}->{element_branch_id} ==
          $inode_cache[$vob_id]->{$file_id}->{element_branch_id}) {

    # if a revision was selected, we start with that revision
    $revision_id = $file_id if $inode_type == iREVI;

    while ($cs_entry->{mkbranch}) {
      my($error);

      $error = mkbranch($view_id, $cs_entry->{mkbranch}, $revision_id,
                        $uid, '');
      return (undef, undef, $error) if $error;
      ($revision_id, $cs_entry) =
          select_revision($view_id, $element_id, $dir_dpath);
      return (undef, undef, KATIEERR_CONFSPECNOCO) if $cs_entry->{nocheckout};
    }
  }

  # unless $version is set, we always checkout .../LATEST
  my($revision_branch) =
      $inode_cache[$vob_id]->{$revision_id}->{element_branch_id};
  my($revision_num) = get_latest_revision_num($vob_db, $revision_branch);
  $revision_id = get_id_by_revision($vob_id, $vob_db,
                                    $revision_branch, $revision_num);
  $branch_id = $inode_cache[$vob_id]->{$revision_id}->{element_branch_id};
  return ($revision_id, $branch_id, undef);
}

sub katie_checkout {
  my($client, $user_info, $view_id,
     $dir_inode_id, $dir_dcookie,
     $name, $reserved, $branch, $ver, $comment) = @_;
  my($vob_id, $vob_db);
  my($dir_dpath);
  my($dir_type, $file_id, $inode_type);
  my($element_id, $branch_id, $revision_id, $cs_entry);
  my($error);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  return NFSERR_INVAL unless defined $comment;
  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $dir_type = get_dir_type($view_id, $dir_inode_id);
  return NFSERR_STALE unless defined($dir_type);

  # can't checkout something from a view-private directory
  return NFSERR_NOTDIR if $dir_type == dVUPR;

  # find the specified file
  $file_id = lookup_name($view_id, $dir_inode_id, $dir_dpath, $name, 1);
  return NFSERR_NOENT unless $file_id;

  # check that the file is not view-private
  $inode_type = get_inode_type($view_id, $file_id);
  return KATIEERR_NOTELEM if $inode_type == iVUPR;

  $element_id = get_element_id($view_id, $file_id);

  # check if it is already checked out to this view
  return KATIEERR_CHECKEDOUT if is_checked_out($view_id, $element_id);

  # check if the config spec lets us checkout the element
  ($revision_id, $cs_entry) =
      select_revision($view_id, $element_id, $dir_dpath);
  return KATIEERR_NOREVSELECTED unless $revision_id;
  return KATIEERR_CONFSPECNOCO if $cs_entry->{nocheckout};

  # must be the VOB owner, the element owner, in the element's group,
  # or root (ensure_group() checks this) to checkout an element
  if ($user_info->{uid} != $vob_cache{$vob_id}->{uid} and
      $user_info->{uid} != $inode_cache[$vob_id]->{$element_id}->{uid} and
      !ensure_group($user_info, $inode_cache[$vob_id]->{$element_id}->{gid})) {
    return NFSERR_PERM;
  }

  ($revision_id, $branch_id, $error) =
      determine_checkout_revision(
          $view_id, $file_id, $element_id, $revision_id, $cs_entry,
          $dir_dpath, $branch, $ver, $user_info->{uid});
  return $error if $error;

  if ($reserved and check_reserved_checkout($vob_id, $vob_db, $revision_id)) {
    return KATIEERR_CHECKEDOUT;
  }

  $name =~ s/\@\@$//;

  # Checking out a regular file
  if ($inode_cache[$vob_id]->{$element_id}->{type} == eREG) {
    my($error, $vupr_file_id, $mode);

    do_checkout($view_id, $element_id, $branch_id, $revision_id,
                $reserved, $user_info->{uid}, $comment, undef);

    # initial mode is the mode of the element, with write permission
    # added where there is currently read permission, and then having
    # the umask subtracted off
    $mode = $inode_cache[$vob_id]->{$element_id}->{mode};
    $mode |= ($mode & 0444) >> 1;
    $mode &= ~$user_info->{umask};

    if ($dir_type != dVIEW) {
      warn "sorry, I don't know where to place the view-private file";
    } else {
      ($error, $vupr_file_id) =
          view_private_create($view_id, $dir_inode_id, eREG, $name,
                              $mode, $user_info->{uid}, $user_info->{gid});

      cluck "unhandled error $error" if ($error);

      $vob_cache{$vob_id}->{store_obj}->construct_revision(
        $file_id, $revision_id, "$view_cache{$view_id}->{store}/$vupr_file_id"
      );
    }

  # Checking out a directory
  } elsif ($inode_cache[$vob_id]->{$file_id}->{type} == eDIR) {
    my($vupr_dir_id);

    $vupr_dir_id =
        view_private_create_co_dir($view_id, $element_id, $revision_id);

    do_checkout($view_id, $element_id, $branch_id, $revision_id,
                $reserved, $user_info->{uid}, $comment, $vupr_dir_id);

  # Unknown file type
  } else {
    return NFSERR_IO;
  }

  dir_mtime_set($view_id, $dir_inode_id);
  touch_element($view_id, $element_id);

  return 0;
}

# TODO: make $preserve_mtime, $save_vupr, $identical, and $from work
sub katie_checkin {
  my($client, $user_info, $view_id, $dir_inode_id, $dir_dcookie,
     $name, $preserve_mtime, $save_vupr, $identical, $from, $comment) = @_;
  my($vob_id, $dir_dpath);
  my($dir_type, $file_id);
  my($checkout_row);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  return NFSERR_INVAL unless defined $comment;
  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if (!$dir_type) {
    return NFSERR_STALE;
  } elsif ($dir_type != dVIEW) {
    # can't checkin something from a non-view directory
    return NFSERR_NOTDIR;
  }

  $name =~ s/\@\@$//;
  $file_id = lookup_name_view($view_id, $dir_inode_id, $dir_dpath, "$name\@\@");
  return NFSERR_NOENT unless $file_id;

  $file_id = get_element_id($view_id, $file_id);
  if (!$file_id) {
    warn "katie_checkin(@_): get_element_id($view_id, $file_id) undef";
    return NFSERR_IO;
  }

  # check that it's checked out
  $checkout_row = is_checked_out($view_id, $file_id);
  if (!$checkout_row) {
    return KATIEERR_NOTCHECKEDOUT;
  }

  if ($inode_cache[$vob_id]->{$file_id}->{type} == eREG) {
    my($vupr_file_id);

    # default $from if it's not already set
    $from = $name unless defined($from);

    # check that there's a view-private regular file under that name
    $vupr_file_id =
        lookup_name_view($view_id, $dir_inode_id, $dir_dpath, $from);
    if (!$vupr_file_id) {
      return NFSERR_NOENT;
    }
    if (get_inode_type($view_id, $vupr_file_id) != iVUPR) {
      return KATIEERR_NOTVUPR;
    }
    if ($inode_cache[$vob_id]->{$vupr_file_id}->{type} != eREG) {
      return KATIEERR_CHECKINNOTREG;
    }

    my($size) = -s "$view_cache{$view_id}->{store}/$vupr_file_id";

    my($error, $new_revision_id) = do_checkin($view_id, $file_id, $size,
                                              $user_info->{uid}, $comment);
    return $error if $error;

    $vob_cache{$vob_id}->{store_obj}->create_revision(
      $file_id, $checkout_row->{revision_id}, $new_revision_id,
      "$view_cache{$view_id}->{store}/$vupr_file_id"
    );

    view_unlink($view_id, $dir_inode_id, $from, $vupr_file_id);
  } elsif ($inode_cache[$vob_id]->{$file_id}->{type} == eDIR) {
    my($vupr_dir_id) = $checkout_row->{view_private_id};

    my($error, $new_revision_id) = do_checkin($view_id, $file_id, 0,
                                              $user_info->{uid}, $comment);
    return $error if $error;

    vob_checkin_dir($view_id, $vupr_dir_id, $new_revision_id);
  } else {
    return NFSERR_IO;
  }

  dir_mtime_set($view_id, $dir_inode_id);
  touch_element($view_id, $file_id);

  return 0;
}

sub katie_mkelem {
  my($client, $user_info,
     $view_id, $dir_inode_id, $dir_dcookie,
     $name, $el_type, $comment) = @_;
  my($vob_id, $dir_dpath);
  my($dir_type);
  my($checkout_row);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  return NFSERR_INVAL unless defined $comment;

  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  # can only create files and directories so far
  if ($el_type != eDIR and
      $el_type != eREG) {
    return KATIEERR_ELTYPEINVAL;
  }

  if (!valid_element_name($name)) {
    return KATIEERR_NAMEINVAL;
  }

  # only users whose primary group is in the VOB group list can create
  # elements
  if (!gid_in_vob_gid_list($vob_id, $user_info->{gid})) {
    return NFSERR_PERM;
  }

  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if (!$dir_type) {
    return NFSERR_STALE;
  } elsif ($dir_type != dVIEW) {
    # can't make an element in a non-view directory
    return NFSERR_NOTDIR;
  }

  # check that it's checked out
  $checkout_row = is_checked_out($view_id, $dir_inode_id);
  if (!$checkout_row) {
    return KATIEERR_NOTCHECKEDOUT;
  }

  # check that an element by the same name doesn't already exist
  if (defined(lookup_name_view($view_id,
                               $dir_inode_id, $dir_dpath, "$name\@\@"))) {
    return NFSERR_EXIST;
  }

  # check that a file by the same name doesn't already exist
  if (defined(lookup_name_view($view_id, $dir_inode_id, $dir_dpath, $name))) {
    return NFSERR_EXIST;
  }

  do_mkelem($vob_id, $view_id, $checkout_row,
            $name, $el_type, $user_info, $comment);
  touch_checkout($view_id, $checkout_row->{id});
  dir_mtime_set($view_id, $dir_inode_id);

  return 0;
}

# XXX -- what if the name being removed is checked out?
sub katie_rmname {
  my($client, $user_info, $view_id, $dir_inode_id, $dir_dcookie, $name,
     $comment) = @_;
  my($vob_id, $dir_dpath);
  my($dir_type);
  my($checkout_row);
  my($file_id, $element_id);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if (!$dir_type) {
    return NFSERR_STALE;
  } elsif ($dir_type != dVIEW) {
    # can't remove a name from a non-view directory
    return NFSERR_NOTDIR;
  }

  # check that it's checked out
  $checkout_row = is_checked_out($view_id, $dir_inode_id);
  if (!$checkout_row) {
    return KATIEERR_NOTCHECKEDOUT;
  }

  # check that the name exists in that directory
  $file_id = lookup_name_view($view_id, $dir_inode_id, $dir_dpath, $name, 1);
  if (!$file_id) {
    return NFSERR_NOENT;
  }

  $element_id = get_element_id($view_id, $file_id);

  if (!$element_id) {
    return KATIEERR_NOTELEM;
  }

  do_rmname($view_id, $checkout_row, $name, $comment);
  touch_checkout($view_id, $checkout_row->{id});
  dir_mtime_set($view_id, $dir_inode_id);

  return 0;
}

sub katie_mklbtype {
  my($client, $user_info, $view_id, $name, $replace, $pbranch) = @_;
  my($vob_id, $vob_db);
  my($name_esc);
  my($row);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  if (!valid_element_name($name) or
      $name eq 'LATEST' or
      $name eq 'CHECKEDOUT' or
      $name =~ /^[0-9]+$/) {
    return KATIEERR_NAMEINVAL;
  }

  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};

  $vob_db->transaction_begin();
  $vob_db->exec("LOCK TABLE branch IN SHARE ROW EXCLUSIVE MODE;");
  $vob_db->exec("LOCK TABLE label_type IN SHARE ROW EXCLUSIVE MODE;");
  $name_esc = $vob_db->escape_sql($name);

  # Check that the label name is not already used as a branch name
  $row = $vob_db->select("
      SELECT COUNT(*)
      FROM   branch
      WHERE  name = '$name_esc';
  ")->fetchrow();
  if ($row->{count} > 0) {
    $vob_db->transaction_commit();
    return KATIEERR_NAMEINVAL;
  }

  # Check if the label name is already used
  # TODO: use get_label_type_id
  $row = $vob_db->select("
      SELECT *
      FROM   label_type
      WHERE  name = '$name_esc';
  ")->fetchrow();

  # If replace is not set and the name is already used, return
  if (!$replace and defined($row)) {
    $vob_db->transaction_commit();
    return KATIEERR_LBTYPEEXIST;
  }

  # If replace is set, the label type already exists, and we are
  # changing from pbranch to non-pbranch, only allow it if there are
  # no instances of this label type
  if ($replace and defined($row) and $row->{pbranch} eq 't' and !$pbranch) {
    my($label_row);

    $label_row = $vob_db->select("
        SELECT COUNT(*)
        FROM   label
        WHERE  label_type_id = $row->{id};
    ")->fetchrow();

    if ($label_row->{count} > 0) {
      return KATIEERR_LESSRESTRICT;
    }
  }

  $pbranch = $pbranch ? "TRUE" : "FALSE";

  # Name already used -- update
  if (defined($row)) {
    $vob_db->update("label_type", {
      pbranch => $pbranch
    }, "id = $row->{id}");

  # Name not used -- insert
  } else {
    $vob_db->insert("label_type", {
      name    => $name,
      uid     => $user_info->{uid},
      gid     => $user_info->{gid},
      pbranch => $pbranch,
    });
  }

  $vob_db->transaction_commit();

  return 0;
}

# TODO: add $version for version-selection
sub katie_mklabel {
  my($client, $user_info,
     $view_id, $replace, $recurse, $label_type,
     $dir_inode_id, $dir_dcookie, $name) = @_;
  my($vob_id, $vob_db, $dir_dpath);
  my($label_type_id, $pbranch);
  my($inode_id, $inode_type);
  my($result);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $vob_db->transaction_begin();
  $vob_db->exec("LOCK TABLE label IN SHARE ROW EXCLUSIVE MODE;");
  $vob_db->exec("LOCK TABLE label_type IN SHARE ROW EXCLUSIVE MODE;");

  $label_type_id = get_label_type_id($vob_db, $label_type);
  if (!$label_type_id) {
    $vob_db->transaction_commit();
    return KATIEERR_LBTYPENEXIST;
  }

  $pbranch = $vob_db->select("
      SELECT pbranch
      FROM   label_type
      WHERE  id = $label_type_id;
  ")->fetchrow()->{pbranch};
  $pbranch = $pbranch eq 't';

  $inode_id = lookup_name($view_id, $dir_inode_id, $dir_dpath, $name, 1);
  if (!$inode_id) {
    $vob_db->transaction_commit();
    return ENOENT;
  }

  $inode_type = get_inode_type($view_id, $inode_id);

  # can't label elements themselves, or branches
  if ($inode_type == iATAT or $inode_type == iELBR) {
    $vob_db->transaction_commit();
    return KATIEERR_NOTAREV;
  }

  if ($inode_type == iVUPR) {
    $vob_db->transaction_commit();
    return KATIEERR_NOTELEM;
  }

  # if an element was specified, get the revision selected by the view
  if ($inode_type == iELEM) {
    $inode_id = select_revision_id($view_id, $inode_id, "$dir_dpath/$name");
    if (!defined($inode_id)) {
      $vob_db->transaction_commit();
      return KATIEERR_NOREVSELECTED;
    }
  }

  $result = mklabel($view_id, $replace, $label_type_id, $pbranch, $inode_id);

  $vob_db->transaction_commit();
  return $result;
}

sub katie_mkbrtype {
  my($client, $user_info, $view_id, $name, $replace, $pbranch, $comment) = @_;
  my($vob_id, $vob_db);
  my($result);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  return NFSERR_INVAL unless defined $comment;

  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};

  if (!valid_element_name($name) or
      $name eq 'LATEST' or
      $name eq 'CHECKEDOUT' or
      $name =~ /^[0-9]+$/) {
    return KATIEERR_NAMEINVAL;
  }

  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};

  $vob_db->transaction_begin();
  $vob_db->exec("LOCK TABLE branch IN SHARE ROW EXCLUSIVE MODE;");
  $vob_db->exec("LOCK TABLE label_type IN SHARE ROW EXCLUSIVE MODE;");

  # create the branch
  $result = mkbrtype($vob_db, $name, $user_info->{uid}, $user_info->{gid},
                     $replace, $pbranch, $comment);

  $vob_db->transaction_commit();

  return $result;
}

# TODO: add $version for version-selection
sub katie_mkbranch {
  my($client, $user_info,
     $view_id, $branch_type,
     $dir_inode_id, $dir_dcookie, $name, $comment) = @_;
  my($vob_id, $vob_db, $dir_dpath);
  my($pbranch);
  my($inode_id, $inode_type);
  my($result);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $vob_db->transaction_begin();
  $vob_db->exec("LOCK TABLE branch IN SHARE ROW EXCLUSIVE MODE;");
  $vob_db->exec("LOCK TABLE element_branch IN SHARE ROW EXCLUSIVE MODE;");

  $inode_id = lookup_name($view_id, $dir_inode_id, $dir_dpath, $name);
  if (!$inode_id) {
    $vob_db->transaction_commit();
    return ENOENT;
  }

  $inode_type = get_inode_type($view_id, $inode_id);

  if ($inode_type == iATAT or $inode_type == iELBR) {
    $vob_db->transaction_commit();
    return KATIEERR_NOTAREV;
  }

  if ($inode_type == iVUPR) {
    $vob_db->transaction_commit();
    return KATIEERR_NOTELEM;
  }

  if ($inode_type == iELEM) {
    $inode_id = select_revision_id($view_id, $inode_id, "$dir_dpath/$name");
    if (!defined($inode_id)) {
      $vob_db->transaction_commit();
      return KATIEERR_NOREVSELECTED;
    }
  }

  $result = mkbranch($view_id, $branch_type, $inode_id,
                     $user_info->{uid}, $comment);

  $vob_db->transaction_commit();
  return $result;
}

sub katie_lsvob {
  my($client, $user_info, @vob_tag_list) = @_;
  my($where) = '';
  my($query, $row);

  # if a vob tag list is specified, figure out a where clause that
  # just gets the tags we're after
  if (@vob_tag_list) {
    $where = "WHERE " .
             join(" OR ",
                  map {
                    "name = '" . $katie_db->escape_sql($_) . "'"
                  } @vob_tag_list
             );
  }

  $query = $katie_db->select("
      SELECT id, name, store
      FROM   vob
      $where;
  ");

  while ($row = $query->fetchrow()) {
    $client->write([{
      id   => $row->{id},
      tag  => $row->{name},
      path => $row->{store},
    }]);
  }

  return undef;
}

sub katie_lsview {
  my($client, $user_info, $vob_id, @view_tag_list) = @_;
  my(@where);
  my($where) = '';
  my($query, $row);

  # if a vob tag list is specified, figure out a where clause that
  # just gets the tags we're after
  if (is_natural($vob_id)) {
    push(@where, "vob_id = $vob_id");
  }
  if (@view_tag_list) {
    push(@where, join(" OR ",
                      map {
                        "view.name = '" . $katie_db->escape_sql($_) . "'"
                      } @view_tag_list
                 ));
  }
  if (@where) {
    $where = "AND " . join(" AND ", map { "($_)" } @where);
  }

  $query = $katie_db->select("
      SELECT view.id, view.name, view.store, vob_id, vob.name AS vob_name
      FROM   view, vob
      WHERE  view.vob_id = vob.id
             $where;
  ");

  while ($row = $query->fetchrow()) {
    $client->write([{
      id      => $row->{id},
      tag     => $row->{name},
      path    => $row->{store},
      vob_id  => $row->{vob_id},
      vob_tag => $row->{vob_name},
    }]);
  }

  return undef;
}

sub gid_in_vob_gid_list {
  my($vob_id, $group) = @_;
  my($row);

  return 1 if $vob_cache{$vob_id}->{gid} == $group;

  $row = $katie_db->select("
      SELECT COUNT(*)
      FROM   vob_group
      WHERE  vob_id = $vob_id
         AND gid = $group;
  ")->fetchrow();

  return $row->{count} > 0;
}

sub katie_protect {
  my($client, $user_info, $view_id,
     $owner, $group, $mode,
     $recurse, $only_file, $only_directory,
     $dir_inode_id, $dir_dcookie, $file) = @_;
  my($vob_id, $dir_dpath);
  my($dir_type);
  my($file_id, $element_id);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $dir_type = get_dir_type($view_id, $dir_inode_id);
  if (!$dir_type) {
    return NFSERR_STALE;
  } elsif ($dir_type == dVUPR) {
    # can't protect anything in a view-private directory
    return NFSERR_NOTDIR;
  }

  ($file_id) = lookup_name($view_id, $dir_inode_id, $dir_dpath, $file);
  if (!$file_id) {
    return NFSERR_NOENT;
  }

  $element_id = get_element_id($view_id, $file_id);

  if (!$element_id) {
    return KATIEERR_NOTELEM;
  }

  # must be root to change owner
  if (defined($owner) and $user_info->{uid} != 0) {
    return NFSERR_PERM;
  }

  # must be root, VOB owner, or element owner to change group or mode
  if ((defined($group) or defined($mode)) and
      $user_info->{uid} != 0 and
      $user_info->{uid} != $vob_cache{$vob_id}->{uid} and
      $user_info->{uid} != $inode_cache[$vob_id]->{$element_id}->{uid}) {
    return NFSERR_PERM;
  }

  # group must be in VOB group list
  if (defined($group) and !gid_in_vob_gid_list($vob_id, $group)) {
    return NFSERR_PERM;
  }

  # must be root or in group being changed to to change group
  if (defined($group) and
      !ensure_group($user_info, $group)) {
    return NFSERR_PERM;
  }

  return do_protect($vob_id, $owner, $group, $mode,
                    $user_info->{umask}, $element_id);
}

sub katie_setcs {
  my($client, $user_info, $view_id, $config_spec) = @_;
  my($vob_id, $vob_db);
  my($error_line);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};

  $error_line = set_cs($vob_db, $view_id, $config_spec);
  return $error_line;
}

sub katie_getcs {
  my($client, $user_info, $view_id) = @_;
  my($vob_id, $vob_db);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};

  return get_cs_entry($vob_db, $view_id)->{spec};
}

sub katie_ls {
  my($client, $user_info, $view_id,
     $want_dir, $dir_inode_id, $dir_dcookie, $file) = @_;
  my($vob_id);
  my($dir_dpath);

  return NFSERR_STALE unless check_vobview_cache($view_id);

  $vob_id = $view_cache{$view_id}->{vob_id};
  $dir_dpath = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $client->write({
    filename => "file $file",
  });
}

sub katie_lshistory {
  my($client, $user_info, $view_id,
     $recurse, $want_dir, $checkout, $branch_name,
     $dir_inode_id, $dir_dcookie, $file) = @_;
  my($vob_id, $vob_db);
  my($file_id, $inode_type);
  my($file_list);
  my($history);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};

  my($dir_dpath) = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $file_id = lookup_name($view_id, $dir_inode_id, $dir_dpath, $file, 1);
  return NFSERR_NOENT unless $file_id;

  # check that the file is an element
  $inode_type = get_inode_type($view_id, $file_id);
  return KATIEERR_NOTELEM unless $inode_type == iELEM;

  my(%file_hash);

  if ($inode_cache[$vob_id]->{$file_id}->{type} == eDIR and !$want_dir) {
    $file_list = get_vob_directory_list($view_id, $dir_inode_id, $dir_dpath,
                                        $recurse);
    for my $file (@$file_list) {
      # ignore '..'
      next if $file->{file_name} =~ /(^|\/)\.\.$/;
#     # ignore '.' and '..'
#     next if $file->{file_name} =~ /(^|\/)\.\.?$/;
      $file_hash{$file->{file_id}} = $file->{file_name};
    }
  } else {
    $file_hash{$file_id} = $file;
  }

  $history = get_object_history($view_id, \%file_hash,
                                $checkout, $branch_name);

  for my $event (@$history) {
    $client->write($event);
  }

  return 0;
}

sub katie_lsvtree {
  my($client, $user_info, $view_id,
     $dir_inode_id, $dir_dcookie, $file, $branch_name,
     $nrecurse, $no_labels, $nco,
     $all_revisions, $all_labels) = @_;
  my($vob_id, $vob_db);
  my($file_id, $inode_type, $element_branch_id);
  my($vtree);

  return NFSERR_STALE unless check_vobview_cache($view_id);
  $vob_id = $view_cache{$view_id}->{vob_id};
  $vob_db = $vob_cache{$vob_id}->{db};

  my($dir_dpath) = get_dpath_by_dcookie($vob_id, $dir_dcookie);

  $file_id = lookup_name($view_id, $dir_inode_id, $dir_dpath, $file, 1);
  return NFSERR_NOENT unless $file_id;

  ($file_id, $element_branch_id) =
      get_elem_and_elbr_id($view_id, $file_id, $branch_name);

  return KATIEERR_NOTELEM unless $file_id;

  $vtree = get_element_vtree($view_id, $file_id, $element_branch_id,
                             $nrecurse, $no_labels, $nco,
                             $all_revisions, $all_labels);

  for my $node (@$vtree) {
    $client->write($node);
  }

  return 0;
}

1;
