# Katie

# Copyright 2000-2001 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::VOB;

use strict;
use Carp qw(cluck);
use Katie::DB;
use Katie::NFSerrno;
use Katie::Errno;
use Katie::Constants;
use Katie::VOBStore;
use Katie::VOBClear;
use Katie::Events;

require Exporter;
@Katie::VOB::ISA = qw(Exporter);

@Katie::VOB::EXPORT = qw(
  %vob_cache

  %view_cache
  @revision_cache
  @inode_type_cache
  @inode_cache
  @dir_mtime_cache

  valid_element_name
  valid_vob_name
  valid_label_name
  vob_exists
  do_mkvob
  do_rmvob
  load_vob_cache
  is_checked_out
  get_main_element_branch
  get_id_by_revision
  get_id_by_label_on_elbr
  get_id_by_label
  get_latest_revision_num
  get_elem_and_elbr_id
  get_element_id
  do_checkout
  do_checkin
  check_reserved_checkout
  do_mkelem
  vob_checkin_dir
  get_label_type_id
  touch_element
  touch_checkout
  mklabel
  do_rmname
  mkbrtype
  mkbranch
  do_protect
  get_object_history
  get_vob_directory_list
  get_element_vtree
);

use vars qw(
  %vob_cache

  %view_cache
  @revision_cache
  @inode_type_cache
  @inode_cache
  @dir_mtime_cache
);

# used to cache the contents of the VOB table
# calling check_vobview_cache() will ensure that the appropriate entry
# is in this table (and the view table)
# $vob_cache{$vob_id}->{}
# additionally, $vob_cache{$vob_id}->{db} is a PGq object
#my(%vob_cache);

# *****
# these things should be moved to View.pm once they can be

# calling check_vobview_cache() will ensure that the appropriate entry
# is in this table (and the vob table)

# used to cache the contents of the VOB and view tables
# $view_cache{$view_id}->{}
#my(%view_cache);

# All these caches are indexed by vob_id (should probably be hashes
# instead of arrays)

# used to cache the inode type (iELEM, ...); this data should
# probably be stored in @inode_cache
# $inode_cache[$vob_id]->{$inode_id}
#my(@inode_type_cache);

# used to cache information about each inode; depending on the
# inode_type, the data will come from one of the following tables:
#   element        -- iELEM, iATAT
#   element_branch -- iELBR
#   revision       -- iREVI
#   view_private   -- iVUPR
# $inode_cache[$vob_id]->{$inode_id}->{data}
#my(@inode_cache);

# used by dir_mtime_get() and dir_mtime_set() to cache the mtimes of
# directories (from the mtime table)
#my(@dir_mtime_cache);

# used by get_id_by_revision()
#my(@revision_cache);

# *****

# FIXME
sub get_inode_type {
  return Katie::View::get_inode_type(@_);
}

sub select_revision_id {
  return Katie::ConfigSpec::select_revision_id(@_);
}

sub get_elbr_id_by_selector {
  return Katie::ConfigSpec::get_elbr_id_by_selector(@_);
}

sub valid_element_name {
  my($name) = @_;

  return ($name ne '.' and
          $name ne '..' and
          substr($name, -2, 2) ne '@@' and
          index($name, '/') == -1 and
          index($name , "\000") == -1);
}

sub valid_vob_name {
  my($name) = @_;

  return ($name =~ /^[a-z][a-z0-9_.-]*$/ and length($name) <= 40);
}

sub valid_label_name {
  local($_) = @_;

  return $_ ne 'CHECKEDOUT' and
         $_ !~ /^[0-9]+$/ and
         $_ !~ /^CHECKEDOUT\.[0-9]+$/ and
         $_ ne '.' and
         $_ ne '..' and
         $_ ne '...';
}

sub vob_exists {
  my($vob_name) = @_;
  my($row);

  $row = $katie_db->select("
      SELECT COUNT(*)
      FROM   vob
      WHERE  name = '$vob_name';
  ")->fetchrow();
  if ($row->{count} > 0) {
    return KATIEERR_VOBEXIST;
  }

  return 0;
}

# create a VOB database and store
# returns ($errno) in case of error or (0, $vob_id) on success
sub do_mkvob {
  my($vob_name, $vob_path, $user_info, $comment) = @_;
  my($vob_row) = {
    name  => $vob_name,
    store => $vob_path,
    uid   => $user_info->{uid},
    gid   => $user_info->{gid},
  };

  if (!mkdir($vob_path, 0700)) {
    return $! + 0;
  }
  if (!mkdir("$vob_path/s", 0700)) {
    return $! + 0;
  }
  if (!mkdir("$vob_path/c", 0700)) {
    return $! + 0;
  }

  $katie_db->insert("vob", $vob_row);
  $vob_row->{id} = $katie_db->get_seq_currval('vob_id_seq');
  $vob_row->{db} = new_vob_db($vob_row->{id});
  $vob_row->{store_obj} =
      Katie::VOBStore->new($vob_row->{store} . "/s");
  $vob_row->{clear_obj} =
      Katie::VOBClear->new($vob_row->{store} . "/c", $vob_row->{store_obj});

  $vob_cache{$vob_row->{id}} = $vob_row;

  for my $gid (@{$user_info->{groups}}) {
    $katie_db->insert("vob_group", {
      vob_id => $vob_row->{id},
      gid    => $gid,
    });
  }

  return (0, $vob_row->{id});
}

# remove a VOB database and store
# assumes all views have been deleted first
sub do_rmvob {
  my($vob_id) = @_;

  # Clear any caches which reference $vob_id
  delete($vob_cache{$vob_id});
  delete($inode_type_cache[$vob_id]);
  delete($inode_cache[$vob_id]);
  delete($dir_mtime_cache[$vob_id]);
  delete($revision_cache[$vob_id]);

  $katie_db->exec("
      DELETE FROM vob
      WHERE  id = $vob_id;
  ");
  $katie_db->exec("
      DELETE FROM vob_group
      WHERE  vob_id = $vob_id;
  ");

  drop_vob_db($vob_id);

  # TODO: write some more code
  # - delete the VOB storage area
}

# given a $vob_id:
# - put the appropriate row from the vob table into %vob_cache
# - create a PGq object to access the vob database, and store it in
#   $vob_cache{$vob_id}->{db}
# return true/false based on whether the $vob_id is valid
sub load_vob_cache {
  my($vob_id) = @_;
  my($query, $row);

  $vob_id = int($vob_id);
  $query = $katie_db->select("
      SELECT *
      FROM   vob
      WHERE  id = $vob_id;
  ");
  $row = $query->fetchrow();
  if (!$row) {
    warn "load_vob_cache($vob_id): unknown vob ID\n";
    return undef;
  }

  $vob_cache{$vob_id} = $row;
  if (!($vob_cache{$vob_id}->{db} = open_vob_db($row->{id}))) {
    return undef;
  }

  $row->{store_obj} =
      Katie::VOBStore->new($row->{store} . "/s");
  $row->{clear_obj} =
      Katie::VOBClear->new($row->{store} . "/c", $row->{store_obj});

  return 1;
}

# given an element_id and a branch_id, return the corresponding inode
# number (which will be of iELBR type)
sub get_main_element_branch {
  my($view_id, $element_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($row);

  $row = $vob_db->select("
      SELECT id
      FROM   element_branch
      WHERE  element_id = $element_id
         AND branch_id = " . MAIN_BRANCH_ID . ";
  ")->fetchrow();
  if (!$row) {
    warn "get_main_element_branch(@_) failed\n";
    return undef;
  }
  return $row->{id};
}

# given an element_branch id and a revision number, return the inode
# number
sub get_id_by_revision {
  my($vob_id, $vob_db, $element_branch_id, $revision_num) = @_;
  my($row);

  if ($revision_cache[$vob_id]->{$element_branch_id}->{$revision_num}) {
    return $revision_cache[$vob_id]->{$element_branch_id}->{$revision_num};
  }

  $row = $vob_db->select("
      SELECT *
      FROM   revision
      WHERE  element_branch_id = $element_branch_id
         AND num = $revision_num;
  ")->fetchrow();

  return undef unless $row;

  $revision_cache[$vob_id]->{$element_branch_id}->{$revision_num} = $row->{id};
  $inode_type_cache[$vob_id]->{$row->{id}} = iREVI;
  $inode_cache[$vob_id]->{$row->{id}} = $row;
  return $row->{id};
}

# given an element_branch_id and a label name, return the
# corresponding revision_id
sub get_id_by_label_on_elbr {
  my($vob_id, $vob_db, $element_branch_id, $label) = @_;
  my($row);

  $label = $vob_db->escape_sql($label);

  $row = $vob_db->select("
      SELECT revision.*
      FROM   revision, label, label_type
      WHERE  revision.element_branch_id = $element_branch_id
         AND revision.id = label.revision_id
         AND label.label_type_id = label_type.id
         AND label_type.name = '$label';
  ")->fetchrow();

  return undef unless $row;

  $inode_type_cache[$vob_id]->{$row->{id}} = iREVI;
  $inode_cache[$vob_id]->{$row->{id}} = $row;
  return $row->{id};
}

# given an element_id and a label name, return the corresponding
# revision_id
sub get_id_by_label {
  my($vob_id, $vob_db, $element_id, $label) = @_;
  my($query, $row);

  $label = $vob_db->escape_sql($label);

  $query = $vob_db->select("
      SELECT revision.*
      FROM   revision, element_branch, label, label_type
      WHERE  revision.element_branch_id = element_branch.id
         AND element_branch.element_id = $element_id
         AND revision.id = label.revision_id
         AND label.label_type_id = label_type.id
         AND label_type.name = '$label';
  ");
  $row = $query->fetchrow();

  return undef unless $row;

  $inode_type_cache[$vob_id]->{$row->{id}} = iREVI;
  $inode_cache[$vob_id]->{$row->{id}} = $row;

  # check for more than two instances of the label on the element
  if ($query->fetchrow()) {
    return "error";
  }

  return $row->{id};
}

# given an inode_id, return the corresponding element's inode_id
# if a valid id is returned, its info will be in the inode_cache
sub get_element_id {
  my($view_id, $inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($inode_type);
  my($element_id);

  if (!($inode_type = get_inode_type($view_id, $inode_id))) {
    return NFSERR_IO;
  }

  if ($inode_type == iELEM) {
    return $inode_id;
  } elsif ($inode_type == iATAT) {
    $element_id = $inode_cache[$vob_id]->{$inode_id}->{id};
  } elsif ($inode_type == iELBR) {
    $element_id = $inode_cache[$vob_id]->{$inode_id}->{element_id};
  } elsif ($inode_type == iREVI) {
    my($el_br_id) = $inode_cache[$vob_id]->{$inode_id}->{element_branch_id};
    if (get_inode_type($view_id, $el_br_id) != iELBR) {
      cluck "get_element_id(@_): inode type of $view_id, $el_br_id not iELBR\n";
      return undef;
    }
    $element_id = $inode_cache[$vob_id]->{$el_br_id}->{element_id};
  } elsif ($inode_type == iCOUT) {
    $element_id = $inode_cache[$vob_id]->{$inode_id}->{element_id};
  } else {
    cluck "get_element_id(@_): get_element_id called with a " .
         "non-element: $inode_id\n";
    return undef;
  }

  # double-check that the pointed-to object is an element
  if (get_inode_type($view_id, $element_id) != iELEM) {
    cluck "get_element_id(@_): inode type of $view_id, $element_id not iELEM\n";
    return undef;
  }
  return $element_id;
}

sub get_latest_revision_num {
  my($vob_db, $element_branch_id) = @_;
  my($row);

  $row = $vob_db->select("
      SELECT max(num)
      FROM   revision
      WHERE  element_branch_id = $element_branch_id;
  ")->fetchrow();

  return undef unless defined($row->{max});
  return $row->{max};
}

# Given an inode_id and an optional branch name, return an element_id
# and an element_branch_id.  The element_id will be the element_id
# corresponding to the inode_id.  The element_branch_id will be the
# element_branch_id of the branch_name (if supplied), otherwise the
# element_branch_id of the inode_id (if of type iELBR, iREVI, or
# iCOUT), otherwise undef.
#
# FIXME: need a way to distinguish between "not an element" error (eg.
# if a VUPR file was passed in) and "bad branch name" error
sub get_elem_and_elbr_id {
  my($view_id, $inode_id, $branch_name) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($inode_type);
  my($elem_id, $elbr_id);

  if (!($inode_type = get_inode_type($view_id, $inode_id))) {
    cluck "get_inode_type returned false\n";
    return undef;
  }

  if ($inode_type == iELEM) {
    $elem_id = $inode_id;
    $elbr_id = undef;

  } elsif ($inode_type == iATAT) {
    $elem_id = $inode_cache[$vob_id]->{$inode_id}->{id};
    $elbr_id = undef;

  } elsif ($inode_type == iELBR) {
    $elem_id = $inode_cache[$vob_id]->{$inode_id}->{element_id};
    $elbr_id = $inode_id;

  } elsif ($inode_type == iREVI) {

    # get the element_branch_id
    $elbr_id = $inode_cache[$vob_id]->{$inode_id}->{element_branch_id};

    # ensure that it is iELBR
    if (!($inode_type = get_inode_type($view_id, $elbr_id))) {
      cluck "get_inode_type on iELBR $inode_id returned false\n";
      return undef;
    }
    if ($inode_type != iELBR) {
      cluck "get_inode_type on iELBR $inode_id returned $inode_type\n";
      return undef;
    }

    $elem_id = $inode_cache[$vob_id]->{$elbr_id}->{element_id};

  } elsif ($inode_type == iCOUT) {

    # get the element_id
    $elem_id = $inode_cache[$vob_id]->{$inode_id}->{element_id};

    # get the revision_id
    my $revision_id = $inode_cache[$vob_id]->{$inode_id}->{revision_id};

    # ensure that it is iREVI
    if (!($inode_type = get_inode_type($view_id, $revision_id))) {
      cluck "get_inode_type on iREVI $inode_id returned false\n";
      return undef;
    }

    if ($inode_type != iREVI) {
      cluck "get_inode_type on iREVI $inode_id returned $inode_type\n";
    }

    # get the element_branch_id
    $elbr_id = $inode_cache[$vob_id]->{$revision_id}->{element_branch_id};
  }

  return ($elem_id, $elbr_id) unless defined $branch_name;

  return undef unless $elbr_id =
      get_elbr_id_by_selector($vob_db, $view_id, $elem_id, $branch_name);

  return ($elem_id, $elbr_id);
}

# XXX to be replaced with some real vob manager stuff
# move the view-private directory into the VOB; the view_private
# directory's contents will be deleted, but the checkout record needs
# to be removed elsewhere
sub vob_checkin_dir {
  my($view_id, $vupr_dir_id, $new_revision_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};

  $vob_db->exec("
      INSERT INTO dir_entry
             (dir_id, element_id, file_name)
      SELECT $new_revision_id, file_id, file_name
      FROM   dir_private_entry
      WHERE  dir_id = $vupr_dir_id;
  ");

  $vob_db->exec("
      DELETE FROM dir_private_entry
      WHERE  dir_id = $vupr_dir_id;
  ");
}

# return the appropriate row from the checkout table if $element_id is
# checked out to $view_id
sub is_checked_out {
  my($view_id, $element_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($query, $row);

  $query = $vob_db->select("
      SELECT *
      FROM   checkout
      WHERE  element_id = $element_id
         AND view_id = $view_id;
  ");
  $row = $query->fetchrow();

  if ($row) {
    $inode_type_cache[$vob_id]->{$row->{id}} = iCOUT;
    return $inode_cache[$vob_id]->{$row->{id}} = $row;
  }

  return undef;
}

# XXX: the checkout table should be locked for writing before calling
# do_checkout()
# Check out revision $revision_id of element $element_id to the view
# $view_id.  For directories only, $view_private_id must be set to the
# inode number of the view-private version of the directory.
# $reserved should be set for a reserved checkout, or undef for an
# unreserved checkout.
# The view-private version of the file/directory must be created
# separately.
sub do_checkout {
  my($view_id, $element_id, $branch_id, $revision_id,
     $reserved, $uid, $comment, $view_private_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($event_id);
  my($row, $id);

  $event_id = create_event($vob_db, $branch_id, 'checkout', $uid, $comment);

  $row = {
    event_id        => $event_id,
    element_id      => $element_id,
    revision_id     => $revision_id,
    view_id         => $view_id,
    reserved        => $reserved ? 'TRUE' : 'FALSE',
    mtime           => time(),
    view_private_id => $view_private_id
  };
  $vob_db->insert("checkout", $row);
  $id = $vob_db->get_seq_currval("obj_id_seq");

  $inode_cache[$vob_id]->{$id} = $row;
  $inode_type_cache[$vob_id]->{$id} = iCOUT;

  return 0;
}

# check in the given element from the given view
#
# XXX should check that the revision that was checked out is currently
# LATEST (if not, the user should do a merge (when a merge is done,
# does the checked out revision change to LATEST?))
#
# the first argument returned is the error number; if 0, the second
# argument contains the inode number (revision ID) of the new version;
# the file/directory data must be created separately
sub do_checkin {
  my($view_id, $element_id, $size, $uid, $comment) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($element_branch_id, $checkout_id, $revision_num);
  my($row);

  # Determine what branch we're checking in on
  $row = $vob_db->select("
      SELECT element_branch_id,
             checkout.id AS checkout_id
      FROM   revision, checkout
      WHERE  checkout.view_id = $view_id
         AND checkout.element_id = $element_id
         AND checkout.revision_id = revision.id
  ")->fetchrow();
  if (!$row) {
    warn "do_checkin couldn't get element_branch_id for @_\n";
    return NFSERR_IO;
  }
  $element_branch_id = $row->{element_branch_id};
  $checkout_id = $row->{checkout_id};

  if (get_inode_type($view_id, $checkout_id) != iCOUT) {
    warn "do_checkin: inode type of $checkout_id was not iCOUT\n";
    return NFSERR_IO;
  }

  # Get the next revision number
  $row = $vob_db->select("
      SELECT MAX(num)
      FROM   revision
      WHERE  element_branch_id = $element_branch_id;
  ")->fetchrow();
  if (!defined($row->{max})) {
    warn "do_checkin couldn't get max(num) for @_\n";
    return NFSERR_IO;
  }
  $revision_num = $row->{max} + 1;

  # Remove the checkout entry
  $vob_db->exec("
      DELETE FROM checkout
      WHERE  element_id = $element_id
         AND view_id = $view_id;
  ");

  destroy_event($vob_db, $inode_cache[$vob_id]->{$checkout_id}->{event_id});

  create_event($vob_db, $checkout_id, "create", $uid, $comment);

  # Create the new revision entry and update the entry in inode_cache
  $vob_db->insert("revision", $inode_cache[$vob_id]->{$checkout_id} = {
    id                => $checkout_id,
    element_branch_id => $element_branch_id,
    num               => $revision_num,
    mtime             => time(),
    size              => $size,
  });
  $inode_type_cache[$vob_id]->{$checkout_id} = iREVI;

  return (0, $checkout_id);
}

sub check_reserved_checkout {
  my($vob_id, $vob_db, $revision_id) = @_;
  my($row);

  # a reserved checkout is wanted -- make sure that noone else has
  # reserved it first
  $row = $vob_db->select("
      SELECT checkout.*
      FROM   checkout, revision r1, revision r2
      WHERE  checkout.revision_id = r1.id
         AND r1.element_branch_id = r2.element_branch_id
         AND r2.id = $revision_id
         AND reserved = TRUE;
  ")->fetchrow();
  if ($row) {
    # may as well store this in case we didn't already know it
    $inode_type_cache[$vob_id]->{$row->{id}} = iCOUT;
    $inode_cache[$vob_id]->{$row->{id}} = $row;

    warn "revision $row->{revision_id} checked out to view $row->{view_id}\n";
    return $row->{view_id};
  }
}

# do_mkelem will create an element:
# $view_id - the view that the directory is checked out to
# $checkout_row - the checked-out directory
# $name - the name of the element to create
# $el_type - eDIR or eREG
#     TODO: make the el_type a string
# note:
# - the parameters must have already been validated
# - if $view_id/$checkout_id/$name are undef, the entry into
#   dir_private_entry won't be made (used to make the VOB root
#   directory), and '..' will point to '.'
sub do_mkelem {
  my($vob_id, $view_id, $checkout_row,
     $name, $el_type, $user_info, $comment) = @_;
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($vob_store) = $vob_cache{$vob_id}->{store};
  my($element_id, $branch_id, $revision_id);

  # create the element
  $vob_db->insert("element", {
    type => $el_type,
    mode => ($el_type == eDIR ? 0777 : 0444) & ~$user_info->{umask},
    uid  => $user_info->{uid},
    gid  => $user_info->{gid},
  });
  # Subtract one from obj_id_seq because we want element.id, not element.id2
  $element_id = $vob_db->get_seq_currval("obj_id_seq") - 1;
  create_event($vob_db, $element_id, "create", $user_info->{uid}, $comment);

  # create the main branch
  $vob_db->insert("element_branch", {
    element_id => $element_id,
    branch_id  => MAIN_BRANCH_ID,
  });
  $branch_id = $vob_db->get_seq_currval("obj_id_seq");
  create_event($vob_db, $branch_id, "create", $user_info->{uid}, '');

  # create revision 0
  $vob_db->insert("revision", {
    element_branch_id => $branch_id,
    num               => 0,
    size              => 0,
  });
  $revision_id = $vob_db->get_seq_currval("obj_id_seq");
  create_event($vob_db, $revision_id, "create", $user_info->{uid}, '');

  # put the directory entry for the new file in the checked out
  # directory
  if (defined($view_id) and defined($checkout_row) and defined($name)) {
    $vob_db->insert("dir_private_entry", {
      view_id   => $view_id,
      dir_id    => $checkout_row->{view_private_id},
      file_id   => $element_id,
      file_name => $name,
    });
    append_event_comment($vob_db, $checkout_row->{event_id},
                         "Added " .
                         ($el_type == eDIR ? "directory" : "file") .
                         " element \"$name\".");
  }

  # for directories, create their initial contents ('.' and '..')
  if ($el_type == eDIR) {
    $vob_db->insert("dir_entry", {
      dir_id     => $revision_id,
      element_id => $element_id,
      file_name  => '.',
    });
    $vob_db->insert("dir_entry", {
      dir_id     => $revision_id,
      element_id => defined($checkout_row)
                    ? $checkout_row->{element_id}
                    : $element_id,
      file_name  => '..',
    });

  # for files, let the VOB store know that there's a new element
  } else {
    $vob_cache{$vob_id}->{store_obj}->create_element(
        $element_id, $branch_id, $revision_id
    );
  }
}

# do_rmname will remove a directory entry from a checked-out directory
# $view_id - the view that the directory is checked out to
# $checkout_row - the checked-out directory
# $name - the name of the element to remove
# the parameters must have already been validated
sub do_rmname {
  my($view_id, $checkout_row, $name, $comment) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};

  $name = $vob_db->escape_sql($name);
  $vob_db->exec("
      DELETE FROM dir_private_entry
      WHERE  view_id = $view_id
         AND dir_id = $checkout_row->{view_private_id}
         AND file_name = '$name';
  ");
  append_event_comment($vob_db, $checkout_row->{event_id},
                       $comment eq ''
                       ? "Removed element \"$name\"."
                       : $comment);
}

sub get_label_type_id {
  my($vob_db, $name) = @_;
  my($row);

  $name = $vob_db->escape_sql($name);
  $row = $vob_db->select("
      SELECT id
      FROM   label_type
      WHERE  name = '$name';
  ")->fetchrow();
  if (!$row) {
    return undef;
  }

  return $row->{id};
}

sub touch_element {
  my($view_id, $element_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($mtime) = time();

  $inode_cache[$vob_id]->{$element_id}->{mtime} = $mtime;

  $vob_db->exec("
      UPDATE element
      SET    mtime = $mtime
      WHERE  id = $element_id;
  ");
}

sub touch_checkout {
  my($view_id, $checkout_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($mtime) = time();

  $inode_cache[$vob_id]->{$checkout_id}->{mtime} = $mtime;

  $vob_db->exec("
      UPDATE checkout
      SET    mtime = $mtime
      WHERE  id = $checkout_id;
  ");
}

sub mklabel {
  my($view_id, $replace, $label_type_id, $pbranch, $revision_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($element_id) = get_element_id($view_id, $revision_id);
  my($where);

  # Select a label that matches $label_type_id on the branch or
  # element as appropriate
  if ($pbranch) {
    # get the element_branch.id that revision_id is on
    my($row) = $vob_db->select("
          SELECT element_branch_id
          FROM   revision
          WHERE  id = $revision_id
        UNION
          SELECT revision.element_branch_id
          FROM   revision, checkout
          WHERE  revision.id = checkout.revision_id
             AND checkout.id = $revision_id
    ")->fetchrow();

    if (!$row) {
      cluck "mklabel(@_) couldn't find element_branch_id\n";
      return NFSERR_IO;
    }

    my($element_branch_id) = $row->{element_branch_id};

    $where = "
        WHERE  label_type_id = $label_type_id
           AND revision_id IN (
                   SELECT revision.id
                   FROM   revision
                   WHERE  revision.element_branch_id = $element_branch_id
                 UNION
                   SELECT checkout.id
                   FROM   checkout, revision
                   WHERE  checkout.revision_id = revision.id
                      AND revision.element_branch_id = $element_branch_id
               )
    ";
  } else {
    $where = "
        WHERE  label_type_id = $label_type_id
           AND revision_id IN (
                   SELECT revision.id
                   FROM   revision, element_branch
                   WHERE  revision.element_branch_id = element_branch.id
                      AND element_branch.element_id = $element_id
                 UNION
                   SELECT checkout.id
                   FROM   checkout
                   WHERE  checkout.element_id = $element_id
               )
    ";
  }

  if ($replace) {
    # Remove any existing label from the branch or element that
    # $revision_id is on
    $vob_db->exec("
        DELETE FROM label
        $where;
    ");
  } else {
    my($row);

    # Check whether the label is already applied to this branch or element
    $row = $vob_db->select("
        SELECT COUNT(*) AS count
        FROM   label, revision, element_branch
        $where;
    ")->fetchrow();

    if ($row->{count} > 0) {
      return KATIEERR_LABELATTACHED;
    }
  }

  $vob_db->insert("label", {
    label_type_id => $label_type_id,
    revision_id   => $revision_id
  });

  touch_element($view_id, $element_id);

  return 0;
}

# FIXME: implement a cache for this (ViewLookup.pm and ViewReaddir.pm
# would probably share it too)
sub get_branch_type_row {
  my($vob_db, $name) = @_;
  my($row);

  $name = $vob_db->escape_sql($name);

  $row = $vob_db->select("
      SELECT id, pbranch
      FROM   branch
      WHERE  name = '$name';
  ")->fetchrow();
  return undef unless $row;

  $row->{pbranch} = $row->{pbranch} eq 't';

  return $row;
}

# precond:
# - the branch name must be valid
# - the branch and label_type tables must be locked in share row
#   exclusive mode
sub mkbrtype {
  my($vob_db, $name, $uid, $gid, $replace, $pbranch, $comment) = @_;
  my($name_esc) = $vob_db->escape_sql($name);
  my($row);

  # Check that the branch name is not already used as a label name
  $row = $vob_db->select("
      SELECT COUNT(*)
      FROM   label_type
      WHERE  name = '$name_esc';
  ")->fetchrow();
  if ($row->{count} > 0) {
    return KATIEERR_NAMEINVAL;
  }

  # Check if the branch name is already used
  $row = $vob_db->select("
      SELECT *
      FROM   branch
      WHERE  name = '$name_esc';
  ")->fetchrow();

  if (!$replace and defined($row)) {
    return KATIEERR_BRTYPEEXIST;
  }

  # If replace is set, the branch type already exists, and we are
  # changing from pbranch to non-pbranch, only allow it if there are
  # no instances of this branch type
  if ($replace and defined($row) and $row->{pbranch} eq 't' and !$pbranch) {
    my($branch_row);

    $branch_row = $vob_db->select("
        SELECT COUNT(*) AS count
        FROM   element_branch
        WHERE  branch_id = $row->{id};
    ")->fetchrow();

    if ($branch_row->{count} > 0) {
      return KATIEERR_LESSRESTRICT;
    }
  }

  $pbranch = $pbranch ? "TRUE" : "FALSE";

  # Name already used -- update
  if (defined($row)) {
    $vob_db->update("branch", {
      pbranch => $pbranch
    }, "id = $row->{id}");
    create_event($vob_db, $row->{id}, 'modify', $uid, $comment);

  # Name not used -- insert
  } else {
    $vob_db->insert("branch", {
      name    => $name,
      uid     => $uid,
      gid     => $gid,
      pbranch => $pbranch,
    });
    my($branch_type_id) = $vob_db->get_seq_currval('obj_id_seq');
    create_event($vob_db, $branch_type_id, 'create', $uid, $comment);
  }

  return 0;
}

sub brtype_exists_on_element {
  my($vob_db, $branch_type_id, $pbranch, $element_id, $revision_id) = @_;
  my(@element_branch_to_check);
  my($query);

  # search for the branch already existing on the branch of $revision_id
  if ($pbranch) {

    $query = "
        SELECT COUNT(*)
        FROM   revision r1, revision r2, element_branch
        WHERE  r1.element_branch_id = r2.element_branch_id
           AND r1.id = $revision_id
           AND r2.id = element_branch.parent_revision_id
           AND element_branch.branch_id = $branch_type_id;
    ";

  # search for the branch already existing on the element of $revision_id
  } else {

    $query = "
        SELECT COUNT(*)
        FROM   element_branch
        WHERE  element_id = $element_id
           AND branch_id = $branch_type_id;
    ";
  }

  if ($vob_db->select($query)->fetchrow()->{count} > 0) {
    return $pbranch ? KATIEERR_BRANCHONBRANCH : KATIEERR_BRANCHONELEM;
  }

  return 0;
}

sub mkbranch {
  my($view_id, $branch_type_name, $revision_id, $uid, $comment) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($branch_type) = get_branch_type_row($vob_db, $branch_type_name);
  my($element_id) = get_element_id($view_id, $revision_id);
  my($error);
  my($new_element_branch_id, $new_revision_id);

  return KATIEERR_BRTYPENEXIST unless $branch_type;

  $error = brtype_exists_on_element($vob_db, $branch_type->{id},
                                    $branch_type->{pbranch},
                                    $element_id, $revision_id);
  return $error if $error;

  $vob_db->insert("element_branch", {
    element_id         => $element_id,
    branch_id          => $branch_type->{id},
    parent_revision_id => $revision_id,
  });

  $new_element_branch_id = $vob_db->get_seq_currval("obj_id_seq");

  create_event($vob_db, $new_element_branch_id, "create", $uid, $comment);

  $vob_db->insert("revision", {
    element_branch_id => $new_element_branch_id,
    num               => 0,
    size              => $inode_cache[$vob_id]->{$revision_id}->{size},
  });

  $new_revision_id = $vob_db->get_seq_currval("obj_id_seq");

  create_event($vob_db, $new_revision_id, "create", $uid, '');

  if ($inode_cache[$vob_id]->{$element_id}->{type} == eREG) {
    $vob_cache{$vob_id}->{store_obj}->create_branch(
        $element_id, $new_element_branch_id, $revision_id, $new_revision_id
    );
  } elsif ($inode_cache[$vob_id]->{$element_id}->{type} == eDIR) {
    $vob_db->exec("
        INSERT INTO dir_entry
               (dir_id, element_id, file_name)
        SELECT $new_revision_id, element_id, file_name
        FROM   dir_entry
        WHERE  dir_id = $revision_id;
    ");
  } else {
    die "mkbranch: type of $element_id is not eREG/eDIR";
  }

  touch_element($view_id, $element_id);

  return 0;
}

# Convert a mode string in either octal or symbolic format into Perl
# code which will make the desired change to $mode when eval'd.
# For example, 'a+r' will be converted into '$mode |= 292;'.
sub mode_to_perl {
  my($mode, $umask) = @_;
  my($need_temp);
  my(@mode_set);

  if ($mode =~ /[0-9]+/) {
    $mode = oct($mode) & 0777;
    return "\$mode = $mode;\n";
  }

  for my $mode_op (split(',', $mode)) {
    my($who, $operator, $permission);
    my($who_mode) = 0;
    my($who_mask);
    my($permission_mode) = 0;
    my($X_permission_mode) = 0;
    my($w_permission_bit);

    if (!(($who, $operator, $permission) =
          ($mode_op =~ /^([ugoa]*)([+-=])([rwxX]*|[ugo])$/))) {
      return undef;
    }

    # $who_mode is a mask with 1 bits in the 'u', 'g', or 'o' positions
    for my $who_char (split(//, $who)) {
      if ($who_char eq 'u') {
        $who_mode |= 0700;
      } elsif ($who_char eq 'g') {
        $who_mode |= 0070;
      } elsif ($who_char eq 'o') {
        $who_mode |= 0007;
      } elsif ($who_char eq 'a') {
        $who_mode |= 0777;
      }
    }
    $who_mask = $who_mode;
    if ($who eq '') {
      $who_mode = 0777 & ~$umask;
    }

    # $permission_mode is a mask with 1 bits in the 'r', 'w', or 'x' positions
    # $X_permission_mode is a mask with 1 bits in the 'x' position
    # $w_permission_bit is the location to get bits from for 'u', 'g', or 'o'
    for my $permission_char (split(//, $permission)) {
      if ($permission_char eq 'r') {
        $permission_mode |= 0444;
      } elsif ($permission_char eq 'w') {
        $permission_mode |= 0222;
      } elsif ($permission_char eq 'x') {
        $permission_mode |= 0111;
      } elsif ($permission_char eq 'X') {
        $X_permission_mode = 0111;
      } elsif ($permission_char eq 'u') {
        $w_permission_bit = 6;
      } elsif ($permission_char eq 'g') {
        $w_permission_bit = 3;
      } elsif ($permission_char eq 'o') {
        $w_permission_bit = 0;
      }
    }

    if (defined($w_permission_bit)) {
      my($mode_mask) = ~(0007 << $w_permission_bit);

      $need_temp = 1;

      push(@mode_set,
           # set $temp_mode to 000X where X are the permissions to be set
           "\$temp_mode = (\$mode & $mode_mask) >> $w_permission_bit;",

           # set $temp_mode to 0XXX
           "\$temp_mode |= (\$temp_mode << 3) | (\$temp_mode << 6);",

           # mask $temp_mode with $who_mode
           "\$temp_mode &= $who_mode;",
      );

      if ($operator eq '+') {
        push(@mode_set, "\$mode |= \$temp_mode;");
      } elsif ($operator eq '-') {
        push(@mode_set, "\$mode &= ~\$temp_mode;");
      } elsif ($operator eq '=') {
        push(@mode_set,
             "\$mode &= ~$who_mask;",
             "\$mode |= \$temp_mode;",
        );
      }
    } else {
      my($combined_mode) = $permission_mode & $who_mode;

      if ($X_permission_mode) {
        $need_temp = 1;
        $X_permission_mode &= $who_mode;
        push(@mode_set,
             # set $temp_mode to $X_permission_mode if any X bits are set
             "\$temp_mode = (\$mode & 0111) ? $X_permission_mode : 0;",

             # add in the rest of $combined_mode to $temp_mode
             "\$temp_mode |= $combined_mode;",
        );
        if ($operator eq '+') {
          push(@mode_set, "\$mode |= \$temp_mode;");
        } elsif ($operator eq '-') {
          push(@mode_set, "\$mode &= ~\$temp_mode;");
        } elsif ($operator eq '=') {
          push(@mode_set,
               "\$mode &= ~$who_mode;",
               "\$mode |= \$temp_mode;",
          );
        }
      } else {
        if ($operator eq '+') {
          push(@mode_set, "\$mode |= $combined_mode;");
        } elsif ($operator eq '-') {
          push(@mode_set, "\$mode &= ~$combined_mode;");
        } elsif ($operator eq '=') {
          push(@mode_set,
               "\$mode &= ~$who_mode;",
               "\$mode |= $combined_mode;",
          );
        }
      }
    }
  }

  if ($need_temp) {
    unshift(@mode_set, 'my($temp_mode);');
  }

  return join("\n", @mode_set) . "\n";
}

# set the owner, group, and/or mode on an element (undef to not set a
# particular item)
# preconditions:
# - valid $element_id, $owner, $group, $umask
# - permission checks have been done
#   - must be root to change owner
#   - must be root, VOB owner, or element owner to change group or mode
#   - must be root or in group to change group
#   - group must be in VOB group list
# returns:
# - NFSERR_IO if invalid $owner or $group
# - KATIEERR_INVALIDMODESTR for bad mode
# - 0 otherwise
sub do_protect {
  my($vob_id, $owner, $group, $mode, $umask, $element_id) = @_;
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($element) = $inode_cache[$vob_id]->{$element_id};
  my($perl_change_mode);
  my($values) = { };

  if (defined($owner)) {
    if ($owner !~ /^[0-9]+$/) {
      warn "in do_protect(@_): bad owner $owner";
      return NFSERR_IO;
    }
    $element->{uid} = $owner;
    $values->{uid} = $owner;
  }

  if (defined($group)) {
    if ($group !~ /^[0-9]+$/) {
      warn "in do_protect(@_): bad group $group";
      return NFSERR_IO;
    }
    $element->{gid} = $group;
    $values->{gid} = $group;
  }

  if (defined($mode)) {
    if (!defined($perl_change_mode = mode_to_perl($mode, $umask))) {
      return KATIEERR_INVALIDMODESTR;
    }

    $mode = $inode_cache[$vob_id]->{$element_id}->{mode};
    eval $perl_change_mode;

    # don't allow the root directory to be chmod'd to 0
    # (otherwise, the owner won't be able to chmod it back)
    if ($mode == 0 and $element_id == ROOT_ELEMENT_ID) {
      $mode = 0100;
    }

    # regular files never get write access
    if ($element->{type} == eREG) {
      $mode &= ~0222;
    }

    $element->{mode} = $mode;
    $values->{mode} = $mode;
  }

  $vob_db->update("element", $values, "id = $element_id");

  return 0;
}

sub get_checkout_row_by_event_id {
  my($view_id, $event_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($row);

  $row = $vob_db->select("
      SELECT *
      FROM   checkout
      WHERE  event_id = $event_id;
  ")->fetchrow();
  return undef unless $row;

  $inode_type_cache[$vob_id]->{$row->{id}} = iCOUT;
  return $inode_cache[$vob_id]->{$row->{id}} = $row;
}

sub get_related_element_list {
  my($vob_db, $checkout, @element_ids) = @_;
  my($element_list);
  my(@query);
  my($query);
  my(%object_hash);

  return {} unless @element_ids;

  $element_list = join(",", @element_ids);

  # checkouts appear on element_branch, so we always need to query it
  push(@query, "
      SELECT element_id, id AS object_id, 'element_branch' AS hint
      FROM   element_branch
      WHERE  element_id IN ($element_list)
  ");

  # unless we only want checkouts, get all other objects
  if (!defined($checkout) or !$checkout) {
    push(@query, "
        SELECT id AS element_id, id AS object_id, 'element' AS hint
        FROM   element
        WHERE  id IN ($element_list)
      UNION
        SELECT element_branch.element_id,
               revision.id AS object_id,
               'revision' AS hint
        FROM   revision, element_branch
        WHERE  revision.element_branch_id = element_branch.id
           AND element_branch.element_id IN ($element_list)
      UNION
        SELECT element_id, id AS object_id, 'checkout' AS hint
        FROM   checkout
        WHERE  element_id IN ($element_list)
    ");
  }
  $query = $vob_db->select(join("UNION\n", @query) . ";");
  {
    my($row);
    $object_hash{$row->{object_id}} = {
      element_id => $row->{element_id},
      hint       => $row->{hint},
    } while $row = $query->fetchrow();
  }

  return \%object_hash;
}

# given a list of object IDs, return the list of all other object IDs
# associated with that object (eg. element, branch, revision,
# checkout) the results are returned as an ref to an array of hashes
# with the element_id and object_id
# precond: filesystem objects must be of type iELEM
sub get_related_object_list {
  my($view_id, $vob_db, $checkout, @object_ids) = @_;
  my(@element_ids, @non_element_ids);
  my($element_list);
  my(@query);
  my($query, $row);
  my($object_hash);

  for my $object_id (@object_ids) {
    my($inode_type) = get_inode_type($view_id, $object_id);
    if (!$inode_type) {
      warn "get_related_object_list called with bad object $object_id\n";
      next;
    } elsif ($inode_type == iELEM) {
      push(@element_ids, $object_id);
    } elsif ($inode_type == iATAT or
             $inode_type == iELBR or
             $inode_type == iREVI) {
      warn "get_related_object_list called with " .
           "non-element filesystem object $object_id\n";
      next;
    } else {
      push(@non_element_ids, $object_id);
    }
  }

  $object_hash = get_related_element_list($vob_db, $checkout, @element_ids);

  for my $object_id (@non_element_ids) {
    $object_hash->{$object_id} = {
      element_id => $object_id,
      hint       => undef,
    };
  }

  return $object_hash;
}

# return the events associated with the given objects
# object_list_hash is a hash ref with object IDs as keys and object
# names as the values
sub get_object_history {
  my($view_id, $object_list_hash, $checkout, $branch_name) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($inode_type);
  my($object_hash, @object_list);
  my($checkout_query);
  my($query, $row);
  my(@event_list);
  my(%branch_list);

  return [] unless keys %$object_list_hash;

  $object_hash = get_related_object_list($view_id, $vob_db, $checkout,
                                         keys %$object_list_hash);

  if (!defined($checkout)) {
    $checkout_query = "";
  } elsif ($checkout) {
    $checkout_query = "AND event.event_type = 'checkout'";
  } else {
    $checkout_query = "AND event.event_type != 'checkout'";
  }

  $query = $vob_db->select("
      SELECT   *
      FROM     event
      WHERE    event.object_id IN (" . (join(",", keys %$object_hash)) . ")
               $checkout_query
      ORDER BY id DESC;
  ");
  while ($row = $query->fetchrow()) {
    $row->{object_type} =
        get_inode_type($view_id, $row->{object_id},
                       $object_hash->{$row->{object_id}}->{hint});
    $row->{version_name} =
        get_version_name_by_object_id($view_id, $vob_id, $row->{object_id});

    if (defined $branch_name) {
      my $branch_list;

      # if we've been given a branch_name, only get objects that can be
      # on a branch
      next unless $row->{object_type} == iELBR ||
                  $row->{object_type} == iREVI ||
                  $row->{object_type} == iCOUT;

      # get the element_id and element_branch_id for this object
      my($element_id, $element_branch_id) =
          get_elem_and_elbr_id($view_id, $row->{object_id});

      # get the list of elbr_ids matching branch_name for this element
      my $elbr_list = exists $branch_list{$element_id}
        ? $branch_list{$element_id}
        : $branch_list{$element_id} =
            [ get_elbr_id_by_selector($vob_db, $view_id,
                                      $element_id, $branch_name) ];

      # check if we have a match
      my $found = 0;
      for (@$elbr_list) {
        $found = 1, last if $_ == $element_branch_id;
      }
      next unless $found;
    }

    if ($row->{event_type} eq 'checkout') {
      my($checkout) = get_checkout_row_by_event_id($view_id, $row->{id});
      if (get_inode_type($view_id, $checkout->{revision_id}) != iREVI) {
        die "get_event_history: revision $checkout->{revision_id} " .
            "in checkout is not, died";
      }
      $row->{pred_version_name} =
          get_version_name_by_object_id($view_id, $vob_id,
                                        $checkout->{revision_id});
      $row->{reserved} = $checkout->{reserved} eq 't';
    }
    $row->{element_id} = $object_hash->{$row->{object_id}}->{element_id};
    $row->{name} = $object_list_hash->{$row->{element_id}};
    push(@event_list, $row)
  }

  return \@event_list;
}

# given an object of type iELBR, return the branch pathname to it (eg.
# '/main' or '/main/mybranch/otherbranch')
sub get_branch_pathname_by_element_branch_id {
  my($view_id, $vob_id, $branch_id) = @_;
  my($inode_cache) = $inode_cache[$vob_id];
  my($branch_type_id, $parent_revision_id);
  my($branch_pathname) = '';

  while (1) {
    # get the leaf branch name
    $branch_type_id = $inode_cache->{$branch_id}->{branch_id};
    if (get_inode_type($view_id, $branch_type_id) != iBRTY) {
      die "get_branch_pathname_by_branch_id(@_) $branch_type_id not iBRTY";
    }
    $branch_pathname =
        "/" .
        $inode_cache->{$branch_type_id}->{name} .
        $branch_pathname;
    $parent_revision_id = $inode_cache->{$branch_id}->{parent_revision_id};

    # check if we're at the top of the branch tree
    return $branch_pathname unless defined $parent_revision_id;

    # we're not, so move down one level
    if (get_inode_type($view_id, $parent_revision_id) != iREVI) {
      die "get_branch_pathname_by_branch_id(@_) $parent_revision_id not iREVI";
    }
    $branch_id = $inode_cache->{$parent_revision_id}->{element_branch_id};
    if (get_inode_type($view_id, $branch_id) != iELBR) {
      die "get_branch_pathname_by_branch_id(@_) $branch_id not iELBR";
    }
  }
}

sub get_version_name_by_revision_id {
  my($view_id, $vob_id, $revision_id) = @_;
  my($branch_id) = $inode_cache[$vob_id]->{$revision_id}->{element_branch_id};
  my($revision_num) = $inode_cache[$vob_id]->{$revision_id}->{num};;

  if (get_inode_type($view_id, $branch_id) != iELBR) {
    die "get_version_name_by_revision_id(@_) $branch_id not iELBR";
  }

  return
      get_branch_pathname_by_element_branch_id($view_id, $vob_id, $branch_id) .
      "/" .
      $revision_num;
}

sub get_version_name_by_object_id {
  my($view_id, $vob_id, $object_id) = @_;
  my($object_type) = $inode_type_cache[$vob_id]->{$object_id};
  my($object) = $inode_cache[$vob_id]->{$object_id};

  return "" if $object_type == iELEM;

  if ($object_type == iELBR) {
    return get_branch_pathname_by_element_branch_id($view_id, $vob_id,
                                                    $object_id);
  } elsif ($object_type == iREVI) {
    return get_version_name_by_revision_id($view_id, $vob_id, $object_id);
  } else {
    die "get_version_name_by_object_id(@_) unknown object type $object_type";
  }
}

sub get_vob_directory_list {
  my($view_id, $dir_inode_id, $dir_dpath, $recurse, $prefix) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($query, $row);
  my(@dir_list);

  $prefix = "" unless defined $prefix;

  my($dir_revision_id) = select_revision_id($view_id, $dir_inode_id,
                                            $dir_dpath);
  if (!$dir_revision_id) {
    warn "get_vob_directory_list: " .
         "select_revision_id(@_) returned undef\n";
    return (NFSERR_IO);
  }

  my($dir_revision_type) = get_inode_type($view_id, $dir_revision_id);

  # Get the contents of the directory from either dir_entry or
  # dir_private_entry depending on whether the directory is checked
  # out or not.
  if ($dir_revision_type == iREVI) {
    $query = $vob_db->select("
        SELECT id AS dir_entry_id, element_id AS file_id, file_name
        FROM   dir_entry
        WHERE  dir_id = $dir_revision_id;
    ");
  } elsif ($dir_revision_type == iCOUT) {
    # get the view private directory ID from the checkout object
    $dir_revision_id =
        $inode_cache[$vob_id]->{$dir_revision_id}->{view_private_id};
    $query = $vob_db->select("
        SELECT id AS dir_entry_id, file_id, file_name
        FROM   dir_private_entry
        WHERE  view_id = $view_id
           AND dir_id = $dir_revision_id;
    ");
  } else {
    warn "get_vob_directory_list(@_): revision returned " .
         "by select_revision_id() ($dir_revision_type) not REVI or COUT\n";
    return (NFSERR_IO);
  }

  while ($row = $query->fetchrow()) {
    if (get_inode_type($view_id, $row->{file_id}) != iELEM) {
      warn "directory $dir_revision_id " .
           "contains non-element object $row->{file_id}\n";
      next;
    }
    my $file_name = $row->{file_name};
    $row->{file_name} = "$prefix/$row->{file_name}" unless $prefix eq '';
    push(@dir_list, $row);
    next unless $recurse;
    next unless $inode_cache[$vob_id]->{$row->{file_id}}->{type} == eDIR;
    next if $file_name eq '.' || $file_name eq '..';
    push(@dir_list,
         @{get_vob_directory_list($view_id, $row->{file_id},
                                  "$dir_dpath/$file_name",
                                  $recurse, "$row->{file_name}")});
  }

  return \@dir_list;
}

# $revision_by_branch is a hash ref keyed by element_branch_id,
# containing an array ref with one element for each revision, which is
# represented by a hash with the following elements:
# - branch_name
# - element_branch_id
# - labels (an array ref containing label names)
# - parent_revision_id - the revision_id of the parent of this
#   revision's branch
# - revision_id
# - revision_num
#
# $branch_parent is a hash ref keyed by revision_id, containing an
# array ref listing the branch_ids that the revision_id is a parent
# for.
#
# $element_branch_id is the element_branch_id indicates the branch to
# start from.
#
# $branch_name is the name of the branch $element_branch_id.
#
# Returns an array of branches and revisions in the order that lsvtree
# expects.  Each of the array elements is hash ref to the particular
# revision from $revision_by_branch, except for the branches, which
# don't have revision_id and revision_num.
#
# The branch_name entry in $revision_by_branch will be changed to be
# the full branch path (eg. "main" to "/main", "foo" to "/main/foo".
#
# Revisions that are branch points or are the last revision on their
# branch will have an "interesting" field set.
sub get_element_vtree_branch {
  my($revision_by_branch, $branch_parent,
     $element_branch_id, $branch_name, $nrecurse) = @_;
  my(@tree);

  my $rev_list = $revision_by_branch->{$element_branch_id};
  my $rev0 = $rev_list->[0];
  my $latest;

  $branch_name .= "/$rev0->{branch_name}";

  # add the branch
  {
    my $branch = {%$rev0};
    delete($branch->{revision_id});
    delete($branch->{revision_num});
    $branch->{branch_name} = $branch_name;
    push(@tree, $branch);
  }

  # add the revisions
  for my $revision (@{$revision_by_branch->{$element_branch_id}}) {
    push(@tree, $revision);
    $revision->{branch_name} = $branch_name;

    if (!$revision->{checkout_id}) {
      $latest = $revision;
    }

    # don't do branches if $nrecurse is set
    next if $nrecurse;

    my $branch_list = $branch_parent->{$revision->{revision_id}};

    # $branch_list will either be undef or be an array with at least
    # one element
    if ($branch_list) {

      # the revision is a branch point -- mark as interesting
      $revision->{interesting} = 1;
      for my $child_element_branch_id (@$branch_list) {
        push(@tree,
             get_element_vtree_branch($revision_by_branch, $branch_parent,
                                      $child_element_branch_id, $branch_name));
      }
    }
  }

  # mark the latest revision on the branch as interesting
  $latest->{interesting} = 1;

  return @tree;
}

sub get_element_vtree {
  my($view_id, $file_id, $element_branch_id,
     $nrecurse, $no_labels, $nco,
     $all_revisions, $all_labels) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($query, $row);
  my(%revision, %revision_by_branch, %branch_parent, $main_element_branch_id);

  {
    my($inode_type) = get_inode_type($view_id, $file_id);
    if (!$inode_type) {
      warn "get_element_vtree called with bad element $file_id\n";
      return [];
    } elsif ($inode_type != iELEM) {
      warn "get_element_vtree called with non-element $file_id\n";
      return [];
    }
  }

  # get all revisions of an element, and put into %revision and
  # %revision_by_branch
  $query = $vob_db->select("
      SELECT   revision.id AS revision_id,
               element_branch.id AS element_branch_id,
               branch.name AS branch_name,
               revision.num AS revision_num,
               element_branch.parent_revision_id,
               NULL AS checkout_id
      FROM     branch, revision, element_branch
      WHERE    revision.element_branch_id = element_branch.id
           AND element_branch.branch_id = branch.id
           AND element_branch.element_id = $file_id
  " . ($nco ? "" : "
    UNION
      SELECT   checkout.revision_id,
               element_branch.id AS element_branch_id,
               branch.name AS branch_name,
               NULL AS revision_num,
               element_branch.parent_revision_id,
               checkout.id AS checkout_id
      FROM     branch, checkout, revision, element_branch
      WHERE    checkout.element_id = $file_id
           AND checkout.revision_id = revision.id
           AND revision.element_branch_id = element_branch.id
           AND element_branch.branch_id = branch.id
  ") . "
    ORDER BY revision_num;
  ");
  while ($row = $query->fetchrow()) {
    $row->{labels} = [];
    $revision{$row->{revision_id}} = $row;
    push(@{$revision_by_branch{$row->{element_branch_id}}}, $row);

    # checkout?
    if (!defined $row->{revision_num}) {
      $row->{revision_num} = "CHECKEDOUT.$row->{checkout_id}";
      next;
    }

    # first revision in a branch?
    if ($row->{revision_num} == 0) {

      # save the parent revision ID for non-main branches
      if ($row->{parent_revision_id}) {
        push(@{$branch_parent{$row->{parent_revision_id}}},
             $row->{element_branch_id});

      # save the branch ID for the main branch so we know where to start
      } else {
        $main_element_branch_id = $row->{element_branch_id};
      }
    }
  }

  unless ($no_labels) {
    # get all labels for the element
    $query = $vob_db->select("
        SELECT   revision_id, name
        FROM     label, label_type, revision, element_branch
        WHERE    label.label_type_id = label_type.id
             AND revision_id = revision.id
             AND revision.element_branch_id = element_branch.id
             AND element_branch.element_id = $file_id
        ORDER BY label_type.id;
    ");
    while ($row = $query->fetchrow()) {
      my $revision = $revision{$row->{revision_id}};
      my $label_list = $revision->{labels};
      if (@$label_list >= 5 && !$all_labels) {
        $revision->{more_labels} = 1;
        next;
      }
      push(@$label_list, $row->{name});
    }
  }

  my @revisions =
    get_element_vtree_branch(
        \%revision_by_branch, \%branch_parent,
        $element_branch_id ? $element_branch_id : $main_element_branch_id,
        "", $nrecurse);

  # return all revisions?
  return [ @revisions ] if $all_revisions;

  # only return the revisions that:
  # - are interesting (branch point or last revision on a branch)
  # - have labels
  # - are checkouts
  # plus the branches
  return [
    grep {
      $_->{interesting} or
      @{$_->{labels}} or
      $_->{checkout_id} or
      !exists $_->{revision_num}
    } @revisions
  ];
}

1;
