# Katie

# Copyright 2000-2002 Geoffrey D. Bennett <geoffrey@netcraft.com.au>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA.

package Katie::View;

use strict;
use Carp qw(cluck);
use Katie::DB;
use Katie::VOB;
use Katie::NFSerrno;
use Katie::ViewFile;
use Katie::Constants;

require Exporter;

@Katie::View::ISA = qw(Exporter);

@Katie::View::EXPORT = qw(
  valid_view_name
  check_vobview_cache
  get_inode_type
  dir_mtime_get
  view_private_create
  view_private_create_co_dir
  view_unlink
  view_rename
  dir_mtime_set
);

# FIXME
sub select_revision_id {
  return Katie::ConfigSpec::select_revision_id(@_);
}

sub valid_view_name {
  my($name) = @_;

  return ($name =~ /^[a-z][a-z0-9_.-]*$/ and length($name) <= 40);
}

# given a $view_id, put the appropriate row from the view table into
# %view_cache
# return true/false based on whether the $view_id is valid
sub load_view_cache {
  my($view_id) = @_;
  my($query, $row);

  $view_id = int($view_id);
  $query = $katie_db->select("
      SELECT *
      FROM   view
      WHERE  id = $view_id;
  ");
  $row = $query->fetchrow();
  if (!$row) {
    warn "load_view_cache($view_id): unknown view ID\n";
    return undef;
  }

  $view_cache{$view_id} = $row;
  return 1;
}

# given a view_id, ensure the appropriate entries are present in
# %view_cache and %vob_cache
# return true/false based on whether the $view_id is valid
sub check_vobview_cache {
  my($view_id) = @_;
  my($vob_id);

  if (!exists($view_cache{$view_id})) {
    if (!load_view_cache($view_id)) {
      return undef;
    }
  }

  $vob_id = $view_cache{$view_id}->{vob_id};

  if (!exists($vob_cache{$vob_id})) {
    if (!load_vob_cache($vob_id)) {
      return undef;
    }
  }

  return 1;
}

# given an object ID, return its type (iELEM, ...)
# postcond: places entries in $inode_type_cache and $inode_cache for
#   $obj_id
# XXX most of this (all but the view-private bit) should be replaced with
# get_object_type() and placed in VOB.pm
sub get_inode_type {
  my($view_id, $obj_id, $hint) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($row);

  if (!defined($obj_id)) {
    cluck "no object ID supplied to get_inode_type";
  }

  if ($inode_type_cache[$vob_id]->{$obj_id}) {
    return $inode_type_cache[$vob_id]->{$obj_id};
  }

  # Check for iELEM, iATAT
  if (!$hint or $hint eq 'element') {
    $row = $vob_db->select("
        SELECT *
        FROM   element
        WHERE  id = $obj_id
            OR id2 = $obj_id;
    ")->fetchrow();
    if ($row) {
      if ($row->{id} == $obj_id) {
        $inode_type_cache[$vob_id]->{$obj_id} = iELEM;
      } else {
        $inode_type_cache[$vob_id]->{$obj_id} = iATAT;
      }
      $inode_cache[$vob_id]->{$row->{id}} = $row;
      $inode_cache[$vob_id]->{$row->{id2}} = $row;
      return $inode_type_cache[$vob_id]->{$obj_id};
    }
  }

  my(@other_object_types) = (
    { type  => iELBR, table => 'element_branch', where => '', },
    { type  => iREVI, table => 'revision',     where => '', },
    { type  => iVUPR, table => 'view_private', where => 'AND link_count > 0' },
    { type  => iCOUT, table => 'checkout',     where => '', },
    { type  => iBRTY, table => 'branch',       where => '', },
  );

  for my $object_type (@other_object_types) {
    if (!$hint or $hint eq $object_type->{table}) {
      $row = $vob_db->select("
          SELECT *
          FROM   $object_type->{table}
          WHERE  id = $obj_id
                 $object_type->{where};
      ")->fetchrow();
      if ($row) {
        $inode_type_cache[$vob_id]->{$obj_id} = $object_type->{type};
        $inode_cache[$vob_id]->{$obj_id} = $row;
        return $object_type->{type};
      }
    }
  }

  cluck "get_inode_type(@_) returning undef";

  return undef;
}

# return the current mtime value of the given directory
sub dir_mtime_get {
  my($view_id, $inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($query, $row);

  if (exists($dir_mtime_cache[$vob_id]->{$view_id}->{$inode_id})) {
    return $dir_mtime_cache[$vob_id]->{$view_id}->{$inode_id};
  }

  $query = $vob_db->select("
      SELECT mtime
      FROM   dir_mtime
      WHERE  view_id = $view_id
         AND dir_id = $inode_id;
  ");
  $row = $query->fetchrow();
  if (!$row) {
    $dir_mtime_cache[$vob_id]->{$view_id}->{$inode_id} = undef;
    return undef;
  }

  $dir_mtime_cache[$vob_id]->{$view_id}->{$inode_id} = $row->{mtime};
}

# set the mtime value of the given directory to the current time
sub dir_mtime_set {
  my($view_id, $inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($time) = time();

  $dir_mtime_cache[$vob_id]->{$view_id}->{$inode_id} = $time;

  if (!$vob_db->update("dir_mtime",
                       { mtime => $time },
                       "view_id = $view_id AND dir_id = $inode_id")) {
    $vob_db->insert("dir_mtime",
                    { view_id => $view_id,
                      dir_id  => $inode_id,
                      mtime   => $time });
  }
}

# XXX to be replaced with some real view manager stuff
# do the common operations that relate to creating a view-private file
# or directory:
# - insert the new file into view_private
# - create the entry in the containing directory
# - touch the containing directory
# for files:
#   - create the file in the view store
# for directories:
#   - create the '.' and '..' entries in the new directory
#   - set the mtime on the new directory
# $view_id - the view to create the file in
# $dir_inode_id - the ID of the containing directory (iELEM or IVUPR)
# $file_type - eREG or eDIR
# $name - the filename to put into the containing directory
# $mode, $uid, $gid - permissions OR
# $existing_inode_id - inode number to link to
#
# If $dir_inode_id is undef, then $name is ignored and the containing
# directory operations are not done (this is useful if we're creating
# the checked-out version of a directory).
#
# If $existing_inode_id is undef, the mode, uid, and gid are used to
# create a new file, but if existing_inode_id is defined, a hard link
# will be created.
sub view_private_create {
  my($view_id, $dir_inode_id, $file_type, $name,
     $mode, $uid, $gid, $existing_inode_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($view_private_data);
  my($inode_id);

  if ($existing_inode_id) {
    $inode_id = $existing_inode_id;

    $vob_db->exec("
        UPDATE view_private
        SET    link_count = link_count + 1
        WHERE  id = $inode_id;
    ");
    $inode_cache[$vob_id]->{$inode_id}->{link_count}++;
  } else {
    $mode &= 07777;
    if ($uid == -1) {
      $uid = 0;
    }
    if ($gid == -1) {
      $gid = 0;
    }

    $view_private_data = {
      link_count => $file_type == eDIR ? 2 : 1,
      type       => $file_type,
      mode       => $mode,
      uid        => $uid,
      gid        => $gid
    };
    $vob_db->insert("view_private", $view_private_data);

    $inode_id = 2147483647 - $vob_db->get_seq_currval("private_id_seq");

    # Put the entry that we just placed in view_private into the cache
    $view_private_data->{id} = $inode_id;
    $inode_cache[$vob_id]->{$inode_id} = $view_private_data;
    $inode_type_cache[$vob_id]->{$inode_id} = iVUPR;
  }

  if (defined($dir_inode_id)) {
    $vob_db->insert("dir_private_entry", {
      view_id   => $view_id,
      dir_id    => $dir_inode_id,
      file_id   => $inode_id,
      file_name => $name
    });

    # Touch the containing directory
    dir_mtime_set($view_id, $dir_inode_id);

    if (!$existing_inode_id) {
      if ($file_type == eREG) {
        view_private_create_file($view_id, $inode_id);
      } else {
        view_private_create_dir($view_id, $inode_id, $dir_inode_id);
      }
    }
  }

  return (0, $inode_id);
}

# XXX to be replaced with some real view manager stuff
# create the given view-private directory
sub view_private_create_dir {
  my($view_id, $inode_id, $dir_inode_id) = @_;
  my($vob_db) = $vob_cache{$view_cache{$view_id}->{vob_id}}->{db};

  $vob_db->insert("dir_private_entry", {
    view_id   => $view_id,
    dir_id    => $inode_id,
    file_id   => $inode_id,
    file_name => '.'
  });
  $vob_db->insert("dir_private_entry", {
    view_id   => $view_id,
    dir_id    => $inode_id,
    file_id   => $dir_inode_id,
    file_name => '..'
  });

  dir_mtime_set($view_id, $inode_id);
}

# XXX to be replaced with some real view manager stuff
# create the given checked-out version of a VOB directory
# $view_id - the view the directory is being checked out to
# $dir_inode_id - the directory (iELEM) that is being checked out
# $revision_id - the particular revision of $dir_inode_id (iREVI)
sub view_private_create_co_dir {
  my($view_id, $dir_inode_id, $revision_id) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};
  my($dir_inode) = $inode_cache[$vob_id]->{$dir_inode_id};
  my($error, $inode_id);

  ($error, $inode_id) =
    view_private_create($view_id, undef, eDIR, undef,
                        $dir_inode->{mode},
                        $dir_inode->{uid}, $dir_inode->{gid});
  if ($error) {
    cluck "unhandled error $error";
  }

  $vob_db->exec("
      INSERT INTO dir_private_entry
             (view_id, dir_id, file_id, file_name)
      SELECT $view_id, $inode_id, element_id, file_name
      FROM   dir_entry
      WHERE  dir_id = $revision_id;
  ");

  dir_mtime_set($view_id, $inode_id);

  return $inode_id;
}

# XXX to be replaced with some real view manager stuff
# remove the given view-private file
sub view_unlink {
  my($view_id, $dir_inode_id, $name, $inode_id) = @_;
  my($view_store) = $view_cache{$view_id}->{store};
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};

  $name = $vob_db->escape_sql($name);
  $vob_db->exec("
      DELETE FROM dir_private_entry
      WHERE  view_id = $view_id
         AND dir_id = $dir_inode_id
         AND file_name = '$name';
  ");

  dir_mtime_set($view_id, $dir_inode_id);

  if ($inode_cache[$vob_id]->{$inode_id}->{link_count} < 1) {
    warn "view_unlink $view_id, $inode_id has link count == 0\n";
    return NFSERR_IO;
  }
  $vob_db->exec("
      UPDATE view_private
      SET    link_count = link_count - 1
      WHERE  id = $inode_id;
  ");
  $inode_cache[$vob_id]->{$inode_id}->{link_count}--;
  if (!$inode_cache[$vob_id]->{$inode_id}->{link_count}) {
    delete($inode_type_cache[$vob_id]->{$inode_id});
    delete($inode_cache[$vob_id]->{$inode_id});
    view_private_unlink($view_id, $inode_id);
  }

  return 0;
}

sub view_rename {
  my($view_id, $old_dir_id, $old_file_name, $new_dir_id, $new_file_name) = @_;
  my($vob_id) = $view_cache{$view_id}->{vob_id};
  my($vob_db) = $vob_cache{$vob_id}->{db};

  $old_file_name = $vob_db->escape_sql($old_file_name);

  $vob_db->update("dir_private_entry", {
    dir_id    => $new_dir_id,
    file_name => $new_file_name
  }, "     view_id   = $view_id" .
     " AND dir_id    = $old_dir_id" .
     " AND file_name = '$old_file_name'"
  );

  return 0;
}

1;
